import slope
import slope.core
from slope.core import (
    Backend,
    Operator,
    Instruction,
    Program,
    MetaOperator,
    ProcedureSet,
    Tensor,
    TensorBuffer,
    SymbolicTensor,
    list_map,
    dtypes,
    devices,
    DType,
    Device,
    CodegenOutput,
)

import numpy as np
from typing import (
    List,
    Any,
)
import os

from slope.operators import operator_set
from slope.procedures import procedure_set
import onnx
import onnxruntime
import tempfile
import pygments
from pygments.lexers.jvm import JavaLexer
from pygments.formatters import Terminal256Formatter


def annotate_shape(symval):
    return f"{symval.shape}, {symval.dtype.name}"


def annotate_sig(in_symvals, out_symvals):
    if isinstance(in_symvals, SymbolicTensor):
        in_symvals = (in_symvals,)
    if isinstance(out_symvals, SymbolicTensor):
        out_symvals = (out_symvals,)
    in_sig = f"({','.join(annotate_shape(t) for t in in_symvals)})"
    out_sig = f"({','.join(annotate_shape(t) for t in out_symvals)})"
    sig = f"{in_sig} -> {out_sig}"
    return sig


class ONNXRuntimeBackend(Backend):
    dtype_for_indices = dtypes.int64
    dtype_map = {
        slope.core.dtypes.float32: "float",
        dtypes.uint8: "uint8",
        dtypes.int8: "int8",
        dtypes.bool: "bool",
        dtypes.int32: "int32",
        dtypes.int64: "int64",
        dtypes.float16: "float16",
    }
    # https://github.com/onnx/onnx/blob/main/onnx/onnx.proto3
    # used for impl args
    onnx_dtype_enum_map = {
        slope.core.dtypes.float32: 1,
        dtypes.uint8: 2,
        dtypes.int8: 3,
        dtypes.int32: 6,
        dtypes.int64: 7,
        dtypes.bool: 9,
        dtypes.float16: 10,
    }

    device_map = {
        devices.cpu: "cpu:0",
        devices.metal: "cpu:1",
        devices.cuda0: "cuda:0",
    }

    target_map = {
        devices.cpu: "CPUExecutionProvider",
        devices.metal: "CoreMLExecutionProvider",
        devices.cuda0: "CUDAExecutionProvider",
    }

    dtype_map_inv = {v: k for k, v in dtype_map.items()}
    device_map_inv = {v: k for k, v in device_map.items()}

    sess_options = onnxruntime.SessionOptions()
    # Disable this flags, easily get nan
    sess_options.execution_mode = onnxruntime.ExecutionMode.ORT_PARALLEL
    # Other flags
    # sess_options.log_severity_level = 3
    # sess_options.use_deterministic_compute = True
    # sess_options.intra_op_num_threads = 1
    sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL
    # sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_EXTENDED
    # sess_options.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_DISABLE_ALL

    def from_numpy(self, val, dtype=None, device=None):
        dtype = dtype or self.DEFAULT_DTYPE
        device = device or self.DEFAULT_DEVICE
        onnx_device = self.device_map[device]
        device_type, device_id = onnx_device.split(":")
        np_val = np.array(val, dtype=dtype.numpy)
        val = onnxruntime.OrtValue.ortvalue_from_numpy(np_val, device_type=device_type, device_id=int(device_id))
        return Tensor(TensorBuffer(val))

    def numpy_of(self, tensor: Tensor, memmap=False):
        if not memmap:
            return tensor.buf.val.numpy()
        with tempfile.NamedTemporaryFile() as arr_file:
            arr = np.memmap(arr_file.name, dtype=tensor.dtype.numpy, shape=tensor.shape, mode="w+")
            arr[:] = tensor.buf.val.numpy()
            return arr

    def shape_of(self, tensor):
        return tuple(tensor.buf.val.shape())

    def dtype_of(self, tensor: Tensor):
        dtype_str = tensor.buf.val.data_type().replace("tensor(", "").replace(")", "")
        return self.dtype_map_inv[dtype_str]

    def device_of(self, tensor: Tensor):
        # TODO: get device idx
        idx = 0 if self.DEFAULT_DEVICE is not devices.metal else 1
        return self.device_map_inv[f"{tensor.buf.val.device_name()}:{idx}"]

    def codegen(self, program: Program, args, *, fn_name: str = "main", fn_defs=dict()) -> List[Any]:
        if fn_name == "main":
            assert not hasattr(self, "fn_count")
            self.fn_count = 0

        def indent(code):
            spaces = " " * (len(code) - len(code.lstrip()))
            spaces += " " * 4
            return "\n".join([spaces + line for line in code.strip().split("\n")])

        # codegen is recursive if jit-of-jit happens
        body_code_lines = []

        for instruction in program.instructions:
            if len(instruction.out_binders) == 0:  # skip codegen for function returns nothing
                continue
            in_vals = list_map(lambda x: program.env[x], instruction.inputs)
            out_vals = list_map(lambda z: program.env[z], instruction.out_binders)
            if isinstance(instruction.op, MetaOperator):
                impl_code, fn_defs = self.impls[instruction.op](args, instruction, fn_defs, in_vals, out_vals)
            else:
                if instruction.op in self.impls.keys():
                    impl_code = self.impls[instruction.op](*in_vals, *out_vals, **instruction.params)
                else:
                    # No impl is defined, fallback to procedure
                    impl_code, fn_defs = self.codegen_impl_as_procedure(args, instruction, fn_defs, in_vals, out_vals)

            for impl_code_line in impl_code.split("\n"):  # handle multi-line code
                body_code_lines += [indent(impl_code_line)]

        in_binders = list_map(lambda x: program.env[x], program.in_binders)
        arg_type_strs = (
            [f"{self.dtype_map[inb.symval.dtype]}[{repr(list(inb.symval.shape))[1:-1]}] {inb.name}" for inb in in_binders]
            if fn_name == "main"
            else [f"{inb.name}" for inb in in_binders]
        )
        fn_args_str = ", ".join(arg_type_strs)

        outs = list_map(lambda x: program.env[x], program.outs)
        out_type_strs = (
            [f"{self.dtype_map[out.symval.dtype]}[{repr(list(out.symval.shape))[1:-1]}] {out.name}" for out in outs]
            if fn_name == "main"
            else [f"{out.name}" for out in outs]
        )
        out_type_str = ", ".join(out_type_strs)

        head_code_lines = []
        if fn_name == "main":
            head_code_lines += ['<ir_version: 7, opset_import: ["" : 18, "slope":1]>']
        head_code_lines += [f"{fn_name} ({fn_args_str}) => ({out_type_str})"]
        model_code_lines = head_code_lines + ["{"] + body_code_lines + ["}"]

        functions_head_def = '<domain: "slope",  opset_import: ["" : 18, "slope":1]>'
        functions_code_lines = []
        for fn_def_code_lines in fn_defs.values():
            functions_code_lines += [functions_head_def] + fn_def_code_lines
        code_lines = model_code_lines + functions_code_lines
        if slope.LOG_JIT:
            formatted_code = pygments.highlight("\n".join(code_lines), JavaLexer(), Terminal256Formatter())
            slope.core.dblog(f"\n---- {program.name} codegen:\n\n" + formatted_code + "\n\n===============\n")

        if fn_name == "main":
            del self.fn_count
        assert len(outs) == len(program.outs)
        return CodegenOutput(
            code_lines=code_lines,
            fn_defs=fn_defs,
            in_binders=in_binders,
            outs=outs,
        )

    def codegen_impl_as_procedure(self, args, instruction: Instruction, fn_defs, in_vals, out_vals):
        op: Operator = instruction.op
        op_name = {v: k for k, v in vars(self.operator_set).items()}[op]
        op_procedure = getattr(self.procedure_set, op_name)
        symvals_in = tuple(inp.symval for inp in instruction.inputs)
        params = instruction.params
        op_program, consts, _ = slope.core.make_program(
            op_procedure,
            *symvals_in,
            static_args=tuple(params.items()),
            name=op.name,
        )
        name = slope.core.jit.get_jit_name(tuple(symvals_in), params)
        if name not in fn_defs.keys():
            op_codegen_output: CodegenOutput = self.codegen(
                op_program,
                args,
                fn_name=name,
                fn_defs=fn_defs,
            )
            fn_defs = {**fn_defs, **op_codegen_output.fn_defs}
            fn_defs[name] = op_codegen_output.code_lines
        in_names = ", ".join(i.name for i in in_vals)
        out_names = ", ".join(o.name for o in out_vals)
        impl_code = f"{out_names} = slope.{name}({in_names})"
        return impl_code, fn_defs

    def compile(self, codegen_output):
        code_lines = codegen_output.code_lines
        code = "\n".join(code_lines)
        model = onnx.parser.parse_model(code)
        device = codegen_output.outs[0].device
        target = self.target_map[device]
        session = onnxruntime.InferenceSession(
            model.SerializeToString(),
            self.sess_options,
            providers=[target],
        )

        def fn(*args):
            io_binding = session.io_binding()
            for a, in_binder in zip(args, codegen_output.in_binders):
                io_binding.bind_input(
                    name=in_binder.name,
                    device_type=a.device_name(),
                    device_id=0,
                    element_type=self.dtype_map_inv[a.data_type().replace("tensor(", "").replace(")", "")].numpy,
                    shape=a.shape(),
                    buffer_ptr=a.data_ptr(),
                )
            for out in codegen_output.outs:
                device_type, device_id = self.device_map[out.device].split(":")
                io_binding.bind_output(out.name, device_type, int(device_id))
            run_options = onnxruntime.RunOptions()
            run_options.log_severity_level = 3
            session.run_with_iobinding(io_binding, run_options)
            outputs = tuple(io_binding.get_outputs())
            return outputs

        return fn, code

    def export(self, jit_output: slope.core.JitOutput, output_path, export_params, input_names, output_names, **kwargs):
        code = jit_output.code
        model = onnx.parser.parse_model(code)
        os.makedirs(output_path, exist_ok=True)
        in_binders = jit_output.codegen_output["in_binders"]
        outs = jit_output.codegen_output["outs"]
        num_consts = jit_output.program.num_consts
        for i in range(num_consts):
            const_array = in_binders[i]["type"].numpy()
            const_name = in_binders[i].name
            const = onnx.numpy_helper.from_array(const_array, name=const_name)
            model.graph.initializer.append(const)

        onnx.save(model.SerializeToString(), os.path.join(output_path, "model.onnx"))
        input_arg_names = [ib.name for ib in in_binders[num_consts:]]
        input_arg_names_str = ", ".join(input_arg_names)
        outs_names = [out.name for out in outs]

        test_input_code = ""
        for i in range(num_consts, len(in_binders)):
            input_name = in_binders[i].name
            input_shape = in_binders[i]["type"].shape
            dtype = in_binders[i]["type"].dtype
            input_dtype = ("np." + dtype.numpy.__name__) if dtype is not dtypes.bool else "bool"
            test_input_code += f"""    {input_name} = np.ones({input_shape}, dtype={input_dtype})\n"""

        module_path = os.path.join(output_path, "__init__.py")
        module_code = f"""import onnxruntime
import os
import numpy as np

root_path = os.path.dirname(__file__)
model_path = os.path.join(root_path, "model.onnx")
session = onnxruntime.InferenceSession(model_path, providers=["CPUExecutionProvider"])
input_arg_names = {input_arg_names}
out_names = {outs_names}

def run(*args, **kwargs):
    if len(args) > 0:
        for a_name, a in zip(input_arg_names, args):
            assert a_name not in kwargs.keys()
            kwargs[a_name] = a
    outputs = session.run(out_names, kwargs)
    return outputs
if __name__ == "__main__":
{test_input_code}
    print("inputs:")
    for inp_name, inp in zip(input_arg_names, ({input_arg_names_str})):
        print(f"{{inp_name}} = ")
        print(inp)
        print(f"dtype: {{inp.dtype}}")
        print(f"shape: {{inp.shape}}")
        print()

    outs = run({input_arg_names_str})

    print("outputs:")
    for out_name, out in zip(out_names, outs):
        print(f"{{out_name}} = ")
        print(out)
        print(f"dtype: {{out.dtype}}")
        print(f"shape: {{out.shape}}")
        print()"""
        with open(module_path, "w") as f:
            f.write(module_code)
            slope.dblog(module_code, enable=slope.LOG_JIT)


backend = ONNXRuntimeBackend(operator_set, procedure_set)


@backend.set_impl(backend.operator_set.jit_op)
def jit_op_impl(self, args, instruction, fn_defs, in_vals, out_vals):
    jit_program = instruction.params["program"]
    jit_name = f"{jit_program.name}"
    if jit_name not in fn_defs.keys():
        jit_codegen_output = self.codegen(
            jit_program,
            args,
            fn_name=jit_name,
            fn_defs=fn_defs,
        )
        fn_defs[jit_name] = jit_codegen_output.code_lines
        fn_defs = {**fn_defs, **jit_codegen_output.fn_defs}
    args_str = ", ".join(i.name for i in in_vals)
    impl_code = f"{', '.join(o.name for o in out_vals)} = slope.{jit_name}({args_str})"
    return impl_code, fn_defs


### Operator Impls


@backend.set_impl(operator_set.cast)
def cast_impl(self, x, y, *, dtype):
    return f"{y.name} = Cast<to={self.onnx_dtype_enum_map[dtype]}>({x.name})"


@backend.set_impl(backend.operator_set.stop_gradient)
def stop_gradient_impl(self, x, y):
    return f"{y.name} = Identity({x.name})"


@backend.set_impl(backend.operator_set.sqrt)
def sqrt_impl(self, x, y):
    return f"{y.name} = Sqrt({x.name})"


@backend.set_impl(backend.operator_set.exp)
def exp_impl(self, x, y):
    return f"{y.name} = Exp({x.name})"


@backend.set_impl(backend.operator_set.log)
def log_impl(self, x, y):
    return f"{y.name} = Log({x.name})"


@backend.set_impl(backend.operator_set.sin)
def sin_impl(self, x, y):
    return f"{y.name} = Sin({x.name})"


@backend.set_impl(backend.operator_set.invert)
def invert_impl(self, x, y):
    return f"{y.name} = Not({x.name})"


@backend.set_impl(backend.operator_set.add)
def add_impl(self, x, w, y):
    return f"{y.name} = Add({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.sub)
def sub_impl(self, x, w, y):
    return f"{y.name} = Sub({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.mul)
def mul_impl(self, x, w, y):
    return f"{y.name} = Mul({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.div)
def div_impl(self, x, w, y):
    return f"{y.name} = Div({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.pow)
def pow_impl(self, x, w, y):
    return f"{y.name} = Pow({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.equal)
def equal_impl(self, x, w, y):
    return f"{y.name} = Equal({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.less)
def less_impl(self, x, w, y):
    return f"{y.name} = Less({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.greater)
def greater_impl(self, x, w, y):
    return f"{y.name} = Greater({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.maximum)
def maximum_impl(self, x, w, y):
    return f"{y.name} = Max({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.matmul)
def matmul_impl(self, x, w, y):
    return f"{y.name} = MatMul({x.name}, {w.name})"


# @backend.set_impl(backend.operator_set.where)
# def where_impl(self, x, w, u, y):
#     return f"""{y.name} = "Where"(%{x.name}, %{w.name}, %{u.name})"""


@backend.set_impl(backend.operator_set.gather_nd)
def gather_nd_impl(self, x, w, y, *, batch_dims):
    return f"{y.name} = GatherND<batch_dims={batch_dims}>({x.name}, {w.name})"


#     return (f"{y.name} = GatherND<batch_dims={batch_dims}>({x.name}, {w.name})"
# if w.symval.dtype is dtypes.int64 else
# f"""{w.name}_ = Cast<to={self.onnx_dtype_enum_map[dtypes.int64]}>({w.name})
# {y.name} = GatherND<batch_dims={batch_dims}>({x.name}, {w.name}_)
# """
# )


@backend.set_impl(backend.operator_set.scatter_nd)
def scatter_nd_impl(
    self,
    x,
    w,
    u,
    y,
):
    return f'{y.name} = ScatterND<reduction="add">({x.name}, {w.name}, {u.name})'


#     if w.symval.dtype is dtypes.int64:
#         return f"{y.name} = ScatterND({x.name}, {w.name}, {u.name})"
#     else:
#         name = f"{w.name}_{random.randrange(100)}"
#         return f"""{name} = Cast<to={self.onnx_dtype_enum_map[dtypes.int64]}>({w.name})
# {y.name} = ScatterND({x.name}, {name}_, {u.name})
# """


@backend.set_impl(operator_set.sum)
def sum_impl(self, x, y, *, dim, keepdim):
    return f"""{y.name}_dim = Constant<value=int64[{len(dim)}]  {{ {repr(dim)[1:(-1 if len(dim) > 1 else -2)]} }} >()
{y.name} = ReduceSum<keepdims={int(keepdim)}> ({x.name}, {y.name}_dim)"""


@backend.set_impl(operator_set.max)
def max_impl(self, x, y, *, dim, keepdim):
    return f"""{y.name}_dim = Constant<value=int64[{len(dim)}]  {{ {repr(dim)[1:(-1 if len(dim) > 1 else -2)]} }} >()
{y.name} = ReduceMax<keepdims={int(keepdim)}> ({x.name}, {y.name}_dim)"""


@backend.set_impl(operator_set.arange)
def arange_impl(self, y, *, start, stop, stride, dtype, device):
    return f"""{y.name}_start = Constant<value_int={start}> ()
{y.name}_limit = Constant<value_int={stop}> ()
{y.name}_delta = Constant<value_int={stride}> ()
{f'''{y.name}_range = Range({y.name}_start, {y.name}_limit, {y.name}_delta)
{y.name} = Cast<to={self.onnx_dtype_enum_map[dtype]}>({y.name}_range)''' if dtype is not dtypes.int64 else
f'''{y.name} = Range({y.name}_start, {y.name}_limit, {y.name}_delta)
'''}"""


# {y.name}_range = Range({y.name}_start, {y.name}_limit, {y.name}_delta)
# {f'{y.name} = Cast<to={self.onnx_dtype_enum_map[dtype]}>({y.name}_range)'}
@backend.set_impl(operator_set.full)
def full_impl(self, y, *, shape, fill_value, dtype, device):
    if dtype is not dtypes.bool:
        if dtypes.is_float(dtype):
            fill_value = float(fill_value)
        elif dtypes.is_int(dtype):
            fill_value = int(fill_value)

        if len(shape) > 0:
            return f"""{y.name}_fill_value=Constant<value={self.dtype_map[dtype]}[1] {{ {fill_value} }}>()
{y.name}_shape = Constant<value=int64[{len(shape)}] {{ {repr(list(shape))[1:-1]} }} >()
{y.name} = Expand({y.name}_fill_value, {y.name}_shape)"""
        else:  # scalar case
            return f"""{y.name}_fill_value=Constant<value={self.dtype_map[dtype]}[1] {{ {fill_value} }}>()
{y.name}_squeeze_dim = Constant<value=int64[1] {{0}}> ()
{y.name} = Squeeze ({y.name}_fill_value, {y.name}_squeeze_dim)"""
    else:
        if len(shape) > 0:
            return f"""{y.name}_fill_value=Constant<value=int64[1] {{{int(fill_value)}}}>()
{y.name}_shape = Constant<value=int64[{len(shape)}] {{ {repr(list(shape))[1:-1]} }} >()
{y.name}_expand = Expand({y.name}_fill_value, {y.name}_shape)
{y.name} = Cast<to={self.onnx_dtype_enum_map[dtype]}>({y.name}_expand)"""
        else:  # scalar case
            return f"""{y.name}_fill_value=Constant<value=int64[1] {{{int(fill_value)}}}>()
{y.name}_squeeze_dim = Constant<value=int64[1] {{0}}> ()
{y.name}_squeeze = Squeeze ({y.name}_fill_value, {y.name}_squeeze_dim)
{y.name} = Cast<to={self.onnx_dtype_enum_map[dtype]}>({y.name}_squeeze)"""


@backend.set_impl(operator_set.random_uniform)
def random_uniform_impl(self, y, *, shape, dtype, device):
    if len(shape) > 0:
        return f"""{y.name} = RandomUniform<dtype={self.onnx_dtype_enum_map[dtype]},shape={repr(list(shape))}>()"""
    else:  # scalar case
        return f"""
{y.name}_rand = RandomUniform<dtype={self.onnx_dtype_enum_map[dtype]}, shape=[1]>()
{y.name}_squeeze_dim = Constant<value=int64[1] {{0}}> ()
{y.name} = Squeeze ({y.name}_rand, {y.name}_squeeze_dim)"""


@backend.set_impl(operator_set.random_normal)
def random_normal_impl(self, y, *, shape, dtype, device):
    if len(shape) > 0:
        return f"""{y.name} = RandomNormal<dtype={self.onnx_dtype_enum_map[dtype]}, shape={repr(list(shape))}>()"""
    else:  # scalar case
        return f"""
{y.name}_randn = RandomNormal<dtype={self.onnx_dtype_enum_map[dtype]}, shape=[1]>()
{y.name}_squeeze_dim = Constant<value=int64[1] {{0}}> ()
{y.name} = Squeeze ({y.name}_randn, {y.name}_squeeze_dim)"""


@backend.set_impl(operator_set.expand)
def expand_impl(self, x, y, *, shape):
    return f"""{y.name}_shape = Constant<value=int64[{len(shape)}] {{ {repr(list(shape))[1:-1]} }} >()
{y.name} = Expand({x.name}, {y.name}_shape)"""


@backend.set_impl(operator_set.reshape)
def reshape_impl(self, x, y, *, shape):
    if len(shape) > 0:
        return f"""{y.name}_shape = Constant<value=int64[{len(shape)}] {{ {repr(list(shape))[1:-1]} }} >()
{y.name} = Reshape({x.name}, {y.name}_shape)"""
    else:  # scalar case
        f"""{y.name}_shape = Constant<value=int64[1] {1} >()
{y.name}_reshape = Reshape({x.name}, {y.name}_shape)
{y.name}_squeeze_dim = Constant<value=int64[1] {{0}}> ()
{y.name} = Squeeze ({y.name}_reshape, {y.name}_squeeze_dim)"""


@backend.set_impl(operator_set.pad)
def pad_impl(self, x, y, *, padding, mode, value):
    padding = padding[::-1]
    padding = padding[0::2] + padding[1::2]
    #     return f"""
    # {y.name}_padding = Constant<value=int64[{len(padding)}]  {{ {repr(list(padding))[1:-1]} }}>()
    # {y.name}_constant_value= Constant<value={value} >()
    # {y.name} = Pad({x.name}, {y.name}_padding, {y.name}_constant_value)
    # """
    return f"""{y.name}_padding = Constant<value=int64[{len(padding)}]  {{ {repr(list(padding))[1:-1]} }}>()
{y.name} = Pad({x.name}, {y.name}_padding)"""


@backend.set_impl(operator_set.slice)
def slice_impl(self, x, y, *, starts, limits, strides):
    return f"""{y.name}_starts = Constant<value=int64[{len(starts)}]  {{ {repr(list(starts))[1:-1]} }}>()
{y.name}_ends = Constant<value=int64[{len(limits)}]  {{ {repr(list(limits))[1:-1]} }}>()
{y.name}_dim = Constant<value=int64[{len(strides)}]  {{ {repr(list(range(len(starts))))[1:-1]} }}>()
{y.name}_steps = Constant<value=int64[{len(strides)}]  {{ {repr(list(strides))[1:-1]} }}>()
{y.name} = Slice({x.name}, {y.name}_starts, {y.name}_ends, {y.name}_dim, {y.name}_steps)"""


@backend.set_impl(operator_set.cat)
def cat_impl(self, *xs, dim):
    xs, y = xs[:-1], xs[-1]
    return f"{y.name} = Concat< axis={dim}>({','.join([x.name for x in xs])})"


@backend.set_impl(operator_set.permute)
def permute_impl(self, x, y, *, perm):
    return f"{y.name} = Transpose<perm={repr(list(perm))}>({x.name})"


@backend.set_impl(operator_set.flip)
def flip_impl(self, x, y, *, dim):
    padding = [0, 0] * x.symval.ndim
    for d in dim:
        padding[d] = 1
    return f"""{x.name}_padding = Constant<value=int64[{len(padding)}]  {{ {repr(list(padding))[1:-1]} }}>()
{x.name}_ = Pad({x.name}, {x.name}_padding)
{y.name}_starts = Constant<value=int64[{len(dim)}]  {{ {", ".join([str(x.symval.shape[d]) for d in dim])}}}>()
{y.name}_ends = Constant<value=int64[{len(dim)}] {{ {", ".join(["0"] * len(dim))}}}>()
{y.name}_dim = Constant<value=int64[{len(dim)}]  {{ {repr(list(dim))[1:-1]} }}>()
{y.name}_steps = Constant<value=int64[{len(dim)}] {{ {", ".join(["-1"] * len(dim))}}}>()
{y.name} = Slice({x.name}_, {y.name}_starts, {y.name}_ends, {y.name}_dim, {y.name}_steps)"""


@backend.set_impl(operator_set.conv)
def conv_impl(self, x, w, y, *, groups, stride, dilation, padding):
    padding = padding[0::2] + padding[1::2]
    dilations_attr = f"dilations=[{repr(list(dilation))[1:-1]}]"
    pads_attr = f"pads=[{repr(list(padding))[1:-1]}]"
    strides_attr = f"strides=[{repr(list(stride))[1:-1]}]"
    group_attr = f"group={groups}"
    return f"""{y.name} = Conv<{dilations_attr}, {pads_attr}, {strides_attr}, {group_attr}>({x.name}, {w.name})"""


# ---- conv_shape__lp_100_cm_64_cm_15_cm_15_rp__dtype_float32_shape__lp_32_cm_64_cm_3_cm_3_rp__dtype_float32_groups_1_stride__lp_1_cm_1_rp__dilation__lp_1_cm_1_rp__padding__lp_1_cm_2_cm_1_cm_2_rp__ codegen:

# func.func @main (%x0: tensor<100x64x15x15xf32>, %x1: tensor<32x64x3x3xf32>) -> (tensor<100x32x16x16xf32>)
# {
#     %y0 = "stablehlo.convolution"(%x0, %x1) {
#       window_strides = dense<[1, 1]> : tensor<2xi64>,
#       padding = dense<[[1, 2], [1, 2]]> : tensor<2x2xi64>,
#       lhs_dilation = dense<1> : tensor<2xi64>,
#       rhs_dilation = dense<[1, 1]> : tensor<2xi64>,
#       window_reversal = dense<false> : tensor<2xi1>,
#       dimension_numbers = #stablehlo.conv<[b, f, 0, 1]x[o, i, 0, 1]->[b, f, 0, 1]>,
#       feature_group_count = 1 : i64,
#       batch_group_count = 1 : i64,
#       precision_config = [#stablehlo<precision DEFAULT>, #stablehlo<precision DEFAULT>]
#     }  : (tensor<100x64x15x15xf32>,tensor<32x64x3x3xf32>) -> (tensor<100x32x16x16xf32>)

#     "func.return"(%y0): (tensor<100x32x16x16xf32>) -> ()
# }

# ===========

# (100, 64, 15, 15) (32, 64, 3, 3) (100, 32, 16, 16)

# ---- conv_shape__lp_100_cm_64_cm_15_cm_15_rp__dtype_float32_shape__lp_32_cm_64_cm_3_cm_3_rp__dtype_float32_groups_1_stride__lp_1_cm_1_rp__dilation__lp_1_cm_1_rp__padding__lp_1_cm_2_cm_1_cm_2_rp__ codegen:

# <ir_version: 7, opset_import: ["" : 18, "slope":1]>
# main (float[100, 64, 15, 15] x0, float[32, 64, 3, 3] x1) => (float[100, 32, 16, 16] y0)
# {
#     y0 = Conv<dilations=[1, 1], pads=[1, 2, 1, 2], strides=[1, 1], group=1>(x0, x1)
# }



import re
import string
from collections import Counter
import numpy as np


def levenshtein(a, b):
    n, m = len(a), len(b)
    if n > m:
        a, b, n, m = b, a, m, n

    current = list(range(n + 1))
    for i in range(1, m + 1):
        previous, current = current, [i] + [0] * n
        for j in range(1, n + 1):
            add, delete = previous[j] + 1, current[j - 1] + 1
            change = previous[j - 1]
            if a[j - 1] != b[i - 1]:
                change = change + 1
            current[j] = min(add, delete, change)

    return current[n]


def word_error_rate(x, y):
    scores = words = 0
    for h, r in zip(x, y):
        h_list = h.split()
        r_list = r.split()
        words += len(r_list)
        scores += levenshtein(h_list, r_list)
    return float(scores) / words, float(scores), words


def one_hot(arr, num_classes=3):
    res = np.eye(num_classes)[np.array(arr).reshape(-1)]
    arr = res.reshape(list(arr.shape) + [num_classes])
    arr = arr.permute((0, 4, 1, 2, 3)).astype(np.float32)
    return arr


def get_dice_score(prediction, target, channel_dim=1, smooth_nr=1e-6, smooth_dr=1e-6):
    channel_dim, reduce_dim = 1, tuple(range(2, len(prediction.shape)))
    prediction = prediction.argmax(dim=channel_dim)
    prediction, target = one_hot(prediction)[:, 1:], one_hot(target)[:, 1:]
    intersection = np.sum(prediction * target, dim=reduce_dim)
    target_sum = np.sum(target, dim=reduce_dim)
    prediction_sum = np.sum(prediction, dim=reduce_dim)
    result = (2.0 * intersection + smooth_nr) / (target_sum + prediction_sum + smooth_dr)
    return result[0]


def normalize_string(s):
    s = "".join(c for c in s.lower() if c not in string.punctuation)
    s = re.sub(r"\b(a|an|the)\b", " ", s)
    return " ".join(s.split())


def f1_score(x, y):
    xt = normalize_string(x).split()
    yt = normalize_string(y).split()
    ct = Counter(xt) & Counter(yt)
    if (ns := sum(ct.values())) == 0:
        return 0.0
    p = ns / len(xt)
    r = ns / len(yt)
    return 2 * p * r / (p + r)



import slope
from slope import nn
import math
def train_loss_fn(w1, w2, x):
    x = x @ w1
    # x = x.exp()
    # x = x.transpose(0,1) @ w2
    loss = x.sum()
    return loss

N = 100
x = slope.randn(N, 784)
w1 = slope.randn(784, 10)
w2 = slope.randn(100, 10)

fn = slope.jit(train_loss_fn)
with slope.core.Timing(f"RUN: "): fn(w1, w2, x)
grad_fn = slope.jit(slope.value_and_grad(train_loss_fn))
grad_fn(w1, w2, x)
print(grad_fn.lower(w1, w2, x).program)
# with slope.core.Timing(f"GRAD: "): grad_fn(w1, w2, x)
with slope.core.Profiling(f"GRAD"): grad_fn(w1, w2, x)
# slope.jit(lambda w, x: slope.jvp(train_loss_fn, (w, x), (w.ones_like(), x.ones_like())))(w, x)
# with slope.core.Profiling(f"RUN"):



import slope


x = slope.tensor([[1, 2], [3, 4]], dtype=slope.float32)
c = x


@slope.jit
def f(x):
    y = (x + c).sum()
    return y


# print(f(x,))
f_jitobj = f.lower(x)
f_jitobj.export("/tmp/f", x)



import slope

from slope import jit, jvp, grad


def f(x):
    aux = x * 2.0
    y = g(aux)
    return y, aux


def g(x):
    y = x.cos() * 2.0
    return y


def f2(x):
    z = x * 2.0
    y = g(z)
    return y


x = slope.tensor(3.0)
x_dot = slope.tensor(1.0)

y1, aux1 = f(x)
y2, aux2 = jit(f)(x)
(y3, _), aux3 = jvp(f, (x,), (x_dot,), has_aux=True)
(y4, _), aux4 = jvp(jit(f), (x,), (x_dot,), has_aux=True)
print(f"{y1=}; {aux1=}\n" f"{y2=}; {aux2=}\n" f"{y3=}; {aux3=}\n" f"{y4=}; {aux4=}\n")
# jacobian (gradient)
gf_x1, aux_g1 = grad(f, has_aux=True)(x)
gf_x2, aux_g2 = grad(jit(f), has_aux=True)(x)
gf_x3, aux_g3 = jit(grad(jit(f), has_aux=True))(x)
(_, gf_x4), aux_g4 = jvp(f, (x,), (x_dot,), has_aux=True)
(_, gf_x5), aux_g5 = jvp(jit(f), (x,), (x_dot,), has_aux=True)
gf_x6, aux_g6 = jit(grad(f, has_aux=True))(x)
print(
    f"{gf_x1=}; {aux_g1=}\n"
    f"{gf_x2=}; {aux_g2=}\n"
    f"{gf_x3=}; {aux_g3=}\n"
    f"{gf_x4=}; {aux_g4=}\n"
    f"{gf_x5=}; {aux_g5=}\n"
    f"{gf_x6=}; {aux_g6=}\n"
)
# hessian (gradient of gradeint)
ggf_x1, aux_gg1 = grad(grad(f, has_aux=True), has_aux=True)(x)
ggf_x2, aux_gg2 = grad(grad(jit(f), has_aux=True), has_aux=True)(x)
ggf_x3, aux_gg3 = grad(jit(grad(f, has_aux=True)), has_aux=True)(x)
ggf_x4, aux_gg4 = jit(grad(grad(f, has_aux=True), has_aux=True))(x)

(_, ggf_x5), aux_gg5 = jvp(grad(f, has_aux=True), (x,), (x_dot,), has_aux=True)
(_, ggf_x6), aux_gg6 = jvp(jit(grad(f, has_aux=True)), (x,), (x_dot,), has_aux=True)
(_, ggf_x7), aux_gg7 = jvp(jit(grad(f, has_aux=True)), (x,), (x_dot,), has_aux=True)
print(
    f"{ggf_x1=}; {aux_gg1=}\n"
    f"{ggf_x2=}; {aux_gg2=}\n"
    f"{ggf_x3=}; {aux_gg3=}\n"
    f"{ggf_x4=}; {aux_gg4=}\n"
    f"{ggf_x5=}; {aux_gg5=}\n"
    f"{ggf_x6=}; {aux_gg6=}\n"
    f"{ggf_x7=}; {aux_gg7=}\n"
)



import unittest

import slope
from slope.core import Tensor
import numpy as np
from typing import NamedTuple


class Result(NamedTuple):
    y: Tensor
    y_dot: Tensor
    L: Tensor
    gL_y: Tensor


class TestOperators(unittest.TestCase):
    @staticmethod
    def run_ad_fns(f, *args, **kwargs):
        args_dot = [slope.ones_like(x) for x in args]
        y, f_lin = slope.linearize(f, *args, **kwargs)
        y_dot = f_lin(*args_dot)
        loss_fn = slope.value_and_grad(lambda *args,: f(*args).sum())
        L, gL_y = loss_fn(*args)
        res = Result(y, y_dot, L, gL_y)
        slope.dblog(f"{args=}", enable=slope.core.backend.LOG_JIT)
        slope.dblog(f"{res=}", enable=slope.core.backend.LOG_JIT)

    def test_maximum(self):
        def f(x, **kwargs):
            z = slope.zeros_like(x)
            out = x.maximum(z)
            return out

        res = self.run_ad_fns(f, slope.tensor([1, 0.5, -0.4, 0, -200]))


if __name__ == "__main__":
    unittest.main()



import slope

XDIMS = (3, 4, 5)
WDIMS = (3, 5, 6)

x = slope.ones(XDIMS)
x_dot = slope.ones(XDIMS)
w = slope.ones(WDIMS)
w_dot = slope.ones(WDIMS)
print(f"{x.shape=}, {w.shape=}")


def f(x, w):
    return x @ w


y = f(x, w)
y_vmap = slope.vmap(f)(x, w)
y_jit_vmap = slope.jit(slope.vmap(f))(x, w)
y_vmap_jit = slope.vmap(slope.jit(f))(x, w)
print(f"{y.shape=}, {y_vmap=} {y_jit_vmap.shape=},  {y_vmap_jit.shape=}")

y_jvp, w_dot_jvp = slope.jvp(f, (x, w), (x_dot, w_dot))
y_jit_jvp, w_dot_jit_jvp = slope.jvp(slope.jit(f), (x, w), (x_dot, w_dot))
print(f"{y_jvp.shape=}, {w_dot_jvp.shape=}")
print(f"{y_jit_jvp.shape=}, {w_dot_jit_jvp.shape=}")

loss_fn = lambda *args: f(*args).sum()
L, (gL_x, gL_w) = slope.value_and_grad(loss_fn, argnums=(0, 1))(x, w)
L_jit, (gL_x_jit, gL_w_jit) = slope.value_and_grad(slope.jit(loss_fn), argnums=(0, 1))(x, w)
print(f"{L.shape=}, {gL_x.shape=}, {gL_w.shape=}")
print(f"{L_jit.shape=}, {gL_x_jit.shape=}, {gL_w_jit.shape=}")



import tensorflow as tf
import tensorflow.mlir.experimental as tme

print('\n#1')
start=tf.constant(0)
limit=tf.constant(10)
delta=tf.constant(2, dtype=tf.int32)
y = tf.range(start, limit, delta); print(f"{y=}")
f =  tme.convert_function((tf.function(tf.range)).get_concrete_function(start, limit, delta))
print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))




import slope
from slope.core import Tensor, SymbolicTensor, TreeDef
from typing import Tuple, List, Optional

import operator as operator_py

from typing import Sequence, Callable, Union, Callable, NamedTuple
import math
import numpy as np

# ====================
# Module
# ====================


class Module:
    def __hash__(self):
        self_flat, treedef = slope.tree_flatten(self)
        return hash(self_flat)

    def __eq__(self, other):
        if type(self) != type(other):
            return False
        return hash(self) == hash(other)

    def get_metadata(self):
        tensor_attrs = dict()  # dict as ordered set
        module_attrs = dict()

        for k, v in self.__dict__.items():
            if isinstance(v, (Tensor, SymbolicTensor)):
                tensor_attrs[k] = None
            elif isinstance(v, (list, tuple)) and not isinstance(v, TreeDef):
                v_flat, v_treedef = slope.tree_flatten(v)
                if all(isinstance(vi, (Tensor, SymbolicTensor)) for vi in v_flat):
                    tensor_attrs[k] = None
            elif isinstance(v, Module):
                module_attrs[k] = None

        static_dict = {k: v for k, v in self.__dict__.items() if k not in tuple(tensor_attrs) + tuple(module_attrs)}
        return dict(
            cls=self.__class__,
            tensor_attrs=tuple(tensor_attrs.keys()),
            module_attrs=tuple(module_attrs.keys()),
            static_dict=static_dict,
        )

    def get_attrs(self, attr_types, with_name=False):
        attrs = dict()
        for k, v in self.__dict__.items():
            if isinstance(v, attr_types):
                attrs[k] = v
        return attrs if with_name else tuple(attrs.values())

    def get_tensors(self, with_name=False):
        return self.get_attrs((Tensor, SymbolicTensor), with_name)

    def get_modules(self, with_name=False):
        return self.get_attrs(Module, with_name)

    def flatten(self):
        metadata = self.get_metadata()
        tensors = tuple(getattr(self, attr) for attr in metadata["tensor_attrs"])
        modules = tuple(getattr(self, attr) for attr in metadata["module_attrs"])
        return metadata, (tensors, modules)

    @staticmethod
    def unflatten(metadata, tensors_modules):
        mod = metadata["cls"].__new__(metadata["cls"])
        mod.__dict__.update(metadata["static_dict"])
        tensors, modules = tensors_modules
        for k, v in zip(metadata["tensor_attrs"], tensors):
            setattr(mod, k, v)
        for k, v in zip(metadata["module_attrs"], modules):
            setattr(mod, k, v)
        return mod

    def leaf_get_metadata(self):
        tensor_attrs = set()
        module_attrs = set()

        def find(obj, prefix):
            nonlocal tensor_attrs, module_attrs
            if isinstance(obj, (Tensor, SymbolicTensor)):
                tensor_attrs.add(prefix.strip("."))
                return
            if isinstance(obj, Module):
                if obj is not self:
                    module_attrs.add(prefix.strip("."))
                for k, v in obj.__dict__.items():
                    find(v, f"{prefix}{str(k)}.")

        find(self, "")
        static_dict = {k: v for k, v in self.__dict__.items() if k not in tuple(tensor_attrs) + tuple(module_attrs)}
        return dict(
            cls=self.__class__,
            tensor_attrs=tuple(tensor_attrs),
            module_attrs=tuple(module_attrs),
            static_dict=static_dict,
        )

    def leaf_flatten(self):
        metadata = self.get_metadata()
        tensors = tuple(operator_py.attrgetter(attr)(self) for attr in metadata["tensor_attrs"])
        rest = dict()
        for mod_attr in metadata["module_attrs"]:
            mod = operator_py.attrgetter(mod_attr)(self)
            mod_rest, _ = mod.flatten()
            rest[mod_attr] = mod_rest

        return (metadata, rest), tensors

    @staticmethod
    def leaf_unflatten(metadata_rest, tensors):
        def reassamble(metadata, rest):
            cls = metadata["cls"]
            mod = cls.__new__(cls)
            mod.__dict__.update(metadata["static_dict"])
            for mod_attr, (metadata_, rest_) in rest.items():
                setattr(mod, mod_attr, reassamble(metadata_, rest_))
            return mod

        metadata, rest = metadata_rest
        mod = reassamble(metadata, rest)

        def set_nested_attr(obj, attr, value):
            nested_attrs = attr.split(".")
            target_obj = obj
            for a in nested_attrs[:-1]:
                target_obj = getattr(target_obj, a)
            setattr(target_obj, nested_attrs[-1], value)

        for tensor, tensor_attr in tuple(zip(tuple(tensors), metadata["tensor_attrs"])):
            set_nested_attr(mod, tensor_attr, tensor)
        return mod

    def export(self, *args, **kwargs):
        f_jitobj = slope.jit(self.__call__).jit_program(*args, **kwargs)
        return


try:
    slope.core.backend.register_node(Module, Module.flatten, Module.unflatten, "Module")
except:
    pass

# ====================
# Optimizers
# ====================


class Optimizer(Module):
    def __init__(self, params, lr: float, filter_fn=None):
        self.params_treedef = slope.tree_flatten(params)[1]
        self.state = Module()
        self.hp = Module()
        self.hp.lr = slope.full((), lr)
        self.iters = slope.zeros((), dtype=slope.int32)
        self.filter_fn = filter_fn
        # TODO: pass a fn to tree_map to get tree binary mask of update/no-update params

    def step(self, p, g, *state_attrs):
        return p, state_attrs

    def __call__(self, params, g_params, nan_to_zero=False):
        state_names, state_attrs = zip(*self.state.get_modules(with_name=True).items())
        if nan_to_zero:
            g_params = slope.tree_map(lambda x: (x == slope.tensor([float("nan")])).where(0.0, x), g_params)
        if self.filter_fn is None:
            step_out, (leaf0, leaf0_treedef) = slope.tree_map(self.step, params, *(g_params, *state_attrs), out_leaf=True)
        # TODO: only update for params updatable True
        else:
            raise NotImplementedError

            def where_step(params, *args):
                def noop(p, *_):
                    return (p, ())

                return slope.where(self.filter_fn, self.step(params, *args), noop(p))

            step_out, (leaf0, leaf0_treedef) = slope.tree_map(where_step, params, *(g_params, *state_attrs), out_leaf=True)

        step_out_T = slope.tree_transpose(self.params_treedef, leaf0_treedef, step_out)
        params_out, state = step_out_T
        # params_out, *state_attrs_out = step_out_T
        # for n, a  in zip(state_names, state_attrs_out):
        #     setattr(self.state, n, a)
        if nan_to_zero:
            params_out = slope.tree_map(lambda x: (x == slope.tensor([float("nan")])).where(0.0, x), params_out)
        self.state = state
        self.iters = self.iters + 1

        return (params_out, self)


class GD(Optimizer):
    def __init__(self, params, lr=0.001):
        super().__init__(params, lr)

    def step(self, p, g, *_):
        lr = self.hp.lr
        p = p - lr * g
        return p, ()


class SGD(Optimizer):
    def __init__(
        self,
        params,
        lr=0.001,
        momentum: float = 0,
        weight_decay=0.0,
        nesterov=False,
    ):
        super().__init__(params, lr)
        self.hp.momentum = momentum
        self.hp.weight_decay = weight_decay
        self.hp.nesterov = nesterov
        self.state.b = slope.tree_map(lambda x: x.zeros_like(), params)

    def step(self, p, g, b):
        lr, m, wd = self.hp.lr, self.hp.momentum, self.hp.weight_decay
        g = g + wd * p
        b = m * b + g
        g = (g + m * b) if self.hp.nesterov else b
        p = p - lr * g
        state = Module()
        state.b = b
        return (p, state)


class Adam(Optimizer):
    def __init__(self, params, lr=0.001, b1=0.9, b2=0.999, eps=1e-5, weight_decay=0.0):
        super().__init__(params, lr)
        self.hp.b1 = b1
        self.hp.b2 = b2
        self.hp.eps = eps
        self.hp.wd = weight_decay
        self.state.m = slope.tree_map(lambda x: x.ones_like(), params)
        self.state.v = slope.tree_map(lambda x: x.ones_like(), params)

    def step(self, p, g, m, v):
        lr, wd = self.hp.lr, self.hp.wd
        b1, b2, eps = self.hp.b1, self.hp.b2, self.hp.eps
        i = self.iters + 1  # slope.ones_like(self.iters)
        m = b1 * m + (1.0 - b1) * g
        v = b2 * v + (1.0 - b2) * (g * g)
        m_hat = m / (1.0 - b1**i)
        v_hat = v / (1.0 - b2**i)
        up = m_hat / ((v_hat).sqrt() + eps)
        if wd > 0:
            up = up + wd * p.stop_gradient()
        p = p - lr * up
        state = Module()
        state.m = m
        state.v = v
        return (p, state)


AdamW = Adam

# class LRScheduler(Module):
#     def __init__(self, optim):
#         self.optim_treedef = slope.tree_flatten(optim)[1]
#         self.iters = slope.zeros(())

#     def step(self, lr):
#         return (lr,)
#         raise NotImplementedError

#     def __call__(self, optim):
#         step_out, (leaf0, leaf0_treedef) = slope.tree_map(self.step, optim.hp.lr, out_leaf=True)
#         step_out_T = slope.tree_transpose(self.optim_treedef, leaf0_treedef, step_out)
#         optim_out = step_out_T
#         self.iters = self.iters + 1
#         return (optim_out, self)


class LRScheduler(Module):
    def __init__(self):
        self.iters = slope.zeros((), dtype=slope.int32)

    def get_lr(self, lr):
        raise NotImplementedError

    def __call__(self, optim):
        optim.hp.lr = self.get_lr(optim.hp.lr)
        self.iters = self.iters + 1
        return (optim, self)


class MultiStepLR(LRScheduler):
    def __init__(self, milestones: List[int], gamma=0.1):
        super().__init__()
        self.milestones = slope.tensor(milestones, dtype=slope.int32)
        self.gamma = slope.tensor(gamma, dtype=slope.float32)

    def get_lr(self, lr) -> Tensor:
        do_step = (self.iters == self.milestones).cast(slope.float32).sum() > 0
        lr_out = slope.where(do_step, lr * self.gamma, lr)
        return lr_out


# class MultiStepLR(LR_Scheduler):
#   def __init__(self, optimizer: Optimizer, milestones: List[int], gamma=0.1):
#     super().__init__(optimizer)
#     self.milestones = milestones
#     self.gamma = gamma

#   def get_lr(self) -> Tensor:
#     if self.epoch_counter.numpy()[0] not in self.milestones:
#       return self.optimizer.lr
#     return self.optimizer.lr * self.gamma

# class ReduceLROnPlateau(LR_Scheduler):
#   def __init__(self, optimizer: Optimizer, mode="min", factor=0.1, patience=10, threshold=1e-4, threshold_mode="rel"):
#     assert mode in ["min", "max"] and threshold_mode in ["rel", "abs"]
#     super().__init__(optimizer)
#     self.mode, self.factor, self.patience, self.threshold, self.threshold_mode = mode, factor, patience, threshold, threshold_mode
#     self.best = float('inf') if mode == "min" else float('-inf')
#     self.bad_epoch = 0

#     if mode == "min": self.threshold *= -1

#   def is_better(self, current: float) -> bool:
#     dynamic_threshold = self.best*(1+self.threshold) if self.threshold_mode == "rel" else self.best+self.threshold
#     if self.mode == "min":
#       return current < dynamic_threshold
#     return current > dynamic_threshold

#   def step(self, current: float) -> None:
#     self.epoch_counter.assign(self.epoch_counter + 1).realize()
#     if self.is_better(current):
#       self.bad_epoch = 0
#       self.best = current
#     else:
#       self.bad_epoch += 1

#     if self.bad_epoch > self.patience:
#       self.optimizer.lr *= self.factor
#       self.bad_epoch = 0

# class CosineAnnealingLR(LR_Scheduler):
#   def __init__(self, optimizer: Optimizer, T_max: int, eta_min=0):
#     super().__init__(optimizer)
#     self.T_max = T_max
#     self.eta_min = eta_min
#     self.eta_max = optimizer.lr.numpy()[0]

#   def get_lr(self) -> Tensor:
#     return Tensor([self.eta_min + 0.5 * (self.eta_max - self.eta_min) * (1 + math.cos((self.epoch_counter.numpy()[0]/self.T_max) * math.pi))], device=self.optimizer.device)

# class OneCycleLR(LR_Scheduler):
#   def __init__(self, optimizer: Optimizer, max_lr: float, div_factor: float, final_div_factor: float, total_steps: int, pct_start: float,
#                anneal_strategy: str = 'linear', cycle_momentum: bool = False):
#     super().__init__(optimizer)
#     self.initial_lr = max_lr / div_factor
#     self.max_lr = max_lr
#     self.min_lr = self.initial_lr / final_div_factor
#     self.total_steps = total_steps
#     self.pct_start = pct_start
#     assert anneal_strategy == 'linear', 'only linear annealing supported'
#     assert not cycle_momentum, 'cycle momentum not supported'
#     self.optimizer.lr.assign(self.get_lr()).realize() # update the initial LR

#   @staticmethod
#   def _annealing_linear(start: float, end: float, pct: Tensor) -> Tensor: return (pct*(end-start)+start)

#   def get_lr(self) -> Tensor:
#     return (self.epoch_counter < self.total_steps*self.pct_start).where(
#       self._annealing_linear(self.initial_lr, self.max_lr, self.epoch_counter/(self.total_steps*self.pct_start)),
#       self._annealing_linear(self.max_lr, self.min_lr, (self.epoch_counter-(self.total_steps*self.pct_start))/(self.total_steps*(1-self.pct_start)))
#     )

# ====================
# Init
# ====================


def compute_fans(shape: Sequence, in_dim=-2, out_dim=-1, batch_dim=()):
    if isinstance(in_dim, int):
        in_size = shape[in_dim]
    else:
        in_size = int(np.prod([shape[i] for i in in_dim]))
    if isinstance(out_dim, int):
        out_size = shape[out_dim]
    else:
        out_size = int(np.prod([shape[i] for i in out_dim]))
    if isinstance(batch_dim, int):
        batch_size = shape[batch_dim]
    else:
        batch_size = int(np.prod([shape[i] for i in batch_dim]))
    receptive_field_size = math.prod(shape) / in_size / out_size / batch_size
    fan_in = in_size * receptive_field_size
    fan_out = out_size * receptive_field_size
    return fan_in, fan_out


def normal(dtype=None) -> Callable:
    def init(shape, dtype=dtype):
        return slope.randn(shape)

    return init


def variance_scaling(
    scale,
    mode: str,
    distribution: str,
    in_dim: Union[int, Sequence[int]] = 1,
    out_dim: Union[int, Sequence[int]] = 0,
    batch_dim: Sequence[int] = (),
    dtype=None,
) -> Callable:
    def init(shape, dtype=dtype):
        fan_in, fan_out = compute_fans(shape, in_dim, out_dim, batch_dim)
        if mode == "fan_in":
            denominator = fan_in
        elif mode == "fan_out":
            denominator = fan_out
        elif mode == "fan_avg":
            denominator = (fan_in + fan_out) / 2
        else:
            raise ValueError(f"invalid mode for variance scaling initializer: {mode}")
        variance = slope.tensor(scale / denominator, dtype=dtype)
        if distribution == "normal":
            return slope.randn(shape) * variance.sqrt()
        elif distribution == "uniform":
            return slope.rand(size=shape.astype(dtype)) * (3 * variance).sqrt()

        else:
            raise ValueError(f"invalid distribution for variance scaling initializer: {distribution}")

    return init


def glorot_normal(
    in_dim: Union[int, Sequence[int]] = 1,
    out_dim: Union[int, Sequence[int]] = 0,
    batch_dim: Sequence[int] = (),
    dtype=None,
) -> Callable:
    return variance_scaling(
        1.0,
        "fan_avg",
        "normal",
        in_dim=in_dim,
        out_dim=out_dim,
        batch_dim=batch_dim,
        dtype=dtype,
    )


def glorot_uniform(
    in_dim: Union[int, Sequence[int]] = 1,
    out_dim: Union[int, Sequence[int]] = 0,
    batch_dim: Sequence[int] = (),
    dtype=None,
) -> Callable:
    return variance_scaling(
        1.0,
        "fan_avg",
        "uniform",
        in_dim=in_dim,
        out_dim=out_dim,
        batch_dim=batch_dim,
        dtype=dtype,
    )


# ====================
# Layers
# ====================


class Fn(Module):
    def __init__(self, fn):
        self.fn = fn

    def __call__(self, x):
        return self.fn(x)


class ReLU(Module):
    def __call__(self, x):
        return x.relu()


class LeakyReLU(Module):
    def __call__(self, x):
        return x.relu()


class Sigmoid(Module):
    def __call__(self, x):
        return x.sigmoid()


class Tanh(Module):
    def __call__(self, x):
        return x.tanh()


class Swish(Module):
    def __call__(self, x):
        return x.swish()


SiLU = Swish


class GELU(Module):
    def __call__(self, x):
        return x.gelu()


class Embedding(Module):
    def __init__(self, vocab_size: int, embed_size: int):
        self.vocab_size = vocab_size
        self.weight = Tensor.glorot_uniform(vocab_size, embed_size)

    def __call__(self, idx: Tensor) -> Tensor:
        if not hasattr(self, "vocab_counter"):
            self.vocab_counter = Tensor.arange(self.vocab_size, requires_grad=False).reshape(1, 1, self.vocab_size)
        return (self.vocab_counter == idx.unsqueeze(2)).expand(*idx.shape, self.vocab_size) @ self.weight


class Linear(Module):
    def __init__(self, in_features, out_features, bias=False):
        self.weight = slope.zeros(out_features, in_features)
        self.bias = slope.zeros(out_features) if bias else None
        self.reset_parameters()

    def __call__(self, x):
        return x.linear(self.weight, self.bias)

    def reset_parameters_(self):
        stdv = 1.0 / math.sqrt(self.weight.shape[1])
        self.weight = slope.tensor(np.random.uniform(-stdv, stdv, self.weight.shape), self.weight.dtype)
        if self.bias is not None:
            self.bias = slope.tensor(np.random.uniform(-stdv, stdv, self.bias.shape), self.bias.dtype)

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.weight.shape[1])
        self.weight = slope.rand_like(self.weight) * 2 * stdv - stdv
        if self.bias is not None:
            self.bias = slope.rand_like(self.bias) * 2 * stdv - stdv


class Sequential(Module):
    def __init__(self, *modules):
        if isinstance(modules[0], (tuple, list)):
            assert len(modules) == 1
            modules = modules[0]
        self.num_modules = len(modules)
        for i in range(len(modules)):
            setattr(self, f"m{i}", modules[i])

    def __call__(self, x, *args, **kwargs):
        modules = [getattr(self, f"m{i}") for i in range(self.num_modules)]
        return x.sequential(modules, *args, **kwargs)


class Pool(Module):
    def __call__(x, kernel_size, stride=1, dilation=1):
        return x.pool(kernel_size, stride, dilation)


class AvgPool2d(Module):
    def __init__(self, kernel_size=2, stride=None):
        self.kernel_size = kernel_size
        self.stride = stride

    def __call__(self, x):
        return x.avgpool2d(self.kernel_size, self.stride)


class MaxPool2d(Module):
    def __call__(x, kernel_size, stride=None):
        return x.maxpool2d(kernel_size, stride)


class ConvNd(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        dims=2,
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.dims = dims
        self.weight = slope.zeros(out_channels, in_channels, *(kernel_size,) * dims)
        self.bias = slope.zeros(out_channels) if bias else None
        self.reset_parameters()

    def __call__(self, x):
        return x.conv(
            self.weight,
            groups=self.groups,
            stride=self.stride,
            dilation=self.dilation,
            padding=self.padding,
        )

    def reset_parameters_(self):
        n = self.in_channels * (self.kernel_size**self.dims)
        stdv = 1.0 / math.sqrt(n)
        self.weight = slope.tensor(np.random.uniform(-stdv, stdv, self.weight.shape), self.weight.dtype)
        if self.bias is not None:
            self.bias = slope.tensor(np.random.uniform(-stdv, stdv, self.bias.shape), self.bias.dtype)

    def reset_parameters(self):
        n = self.in_channels * (self.kernel_size**self.dims)
        stdv = 1.0 / math.sqrt(n)
        self.weight = slope.rand_like(self.weight) * 2 * stdv - stdv
        if self.bias is not None:
            self.bias = slope.rand_like(self.bias) * 2 * stdv - stdv


class Conv1d(ConvNd):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            dims=1,
        )


class Conv2d(ConvNd):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
            dims=2,
        )


class ConvNdTranspose(Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
        dims=2,
        output_padding=0,
        W_init=glorot_normal(),
    ):
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.dims = dims
        if output_padding != 0:
            raise NotImplementedError
        self.output_padding = output_padding
        self.weight = W_init((out_channels, in_channels, *(kernel_size,) * dims))
        self.bias = Tensor.zeros(out_channels) if bias else None

    def __call__(self, x):
        return x.conv_transpose(
            self.weight,
            groups=self.groups,
            stride=self.stride,
            dilation=self.dilation,
            padding=self.padding,
            output_padding=self.output_padding,
        )


# ==============
# Regularization
# =============


class BatchNorm(Module):
    def __init__(self, num_features: int, eps=1e-5, affine=True, momentum=0.1):
        self.eps = eps
        self.momentum = momentum

        self.weight = slope.ones(num_features) if affine else None
        self.bias = slope.zeros(num_features) if affine else None

        self.running_mean = slope.zeros(num_features)
        self.running_var = slope.ones(num_features)
        self.num_batches_tracked = slope.zeros((), dtype=slope.int32)

    def __call__(self, x, training=True, track_running_stats=True):
        if training:
            D = len(x.shape[2:])
            broadcast_shape = (1, -1) + (1,) * D
            reduce_dim = (0,) + tuple(2 + i for i in range(D))
            mean = x.mean(reduce_dim)
            z = x - mean.reshape(broadcast_shape)
            var = (z * z).mean(reduce_dim)
            invstd = (var + self.eps).rsqrt()
            if track_running_stats:
                z_numel = math.prod(z.shape)
                z_ratio = (z_numel / (z_numel - z.shape[1])) if z_numel != z.shape[1] else 1
                self.running_mean = (1 - self.momentum) * self.running_mean + self.momentum * mean
                self.running_var = (1 - self.momentum) * self.running_var + self.momentum * z_ratio * var
                self.num_batches_tracked = self.num_batches_tracked + 1
        else:
            mean = self.running_mean
            invstd = (self.running_var + self.eps).rsqrt()

        return x.batchnorm(self.weight, self.bias, mean, invstd)


BatchNorm1d = BatchNorm2d = BatchNorm


class LayerNorm(Module):
    def __init__(
        self,
        normalized_shape: Union[int, Tuple[int, ...]],
        eps: float = 1e-5,
        elementwise_affine: bool = True,
    ):
        self.normalized_shape = (normalized_shape,) if isinstance(normalized_shape, int) else tuple(normalized_shape)
        self.dim, self.eps, self.elementwise_affine = (
            tuple(-1 - i for i in range(len(self.normalized_shape))),
            eps,
            elementwise_affine,
        )
        self.weight, self.bias = (
            (
                Tensor.ones(*self.normalized_shape),
                Tensor.zeros(*self.normalized_shape),
            )
            if elementwise_affine
            else (None, None)
        )

    def __call__(self, x: Tensor):
        assert (
            self.normalized_shape == x.shape[-len(self.normalized_shape) :]
        ), f"last dimensions of {x.shape} must match {self.normalized_shape}"
        x = x.layernorm(eps=self.eps, dim=self.dim)
        if not self.elementwise_affine:
            return x
        return x * self.weight + self.bias


class LayerNorm1d(LayerNorm):
    def __call__(self, x):
        return super().__call__(x.transpose(1, 2)).transpose(1, 2)


class LayerNorm2d(LayerNorm):
    def __call__(self, x):
        return super().__call__(x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)


class Dropout(Module):
    def __init__(self, p: float):
        self.p = p

    def __call__(self, x, training=False):
        return x.dropout(self.p, training)


class ScaledDotProductAttention(Module):
    def __call__(
        x,
        key: Tensor,
        value: Tensor,
        attn_mask: Optional[Tensor] = None,
        dropout_p: float = 0.0,
        is_causal: bool = False,
    ):
        return x.scaled_dot_product_attention(key, value, attn_mask, dropout_p, is_causal)


class BCELoss(Module):
    def __call__(x, y):
        return x.binary_cross_entropy(y)


class BCEWithLogitsLoss(Module):
    def __call__(x, y):
        return x.binary_cross_entropy_with_logits(y)


class CrossEntropyLoss(Module):
    def __call__(x, y):
        return x.cross_entropy(y)


class Softmax(Module):
    def __call__(x, dim=-1):
        return x.softmax(dim)


class LogSoftmax(Module):
    def __call__(x, dim=-1):
        return x.log_softmax(dim)



# A set of code samples showing different usage of the ONNX Runtime Python API
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnxruntime
import onnx
# from onnx import parser, printer

MODEL_FILE = '.model.onnx'
DEVICE_NAME = 'cpu'
DEVICE_INDEX = 0
DEVICE=f'{DEVICE_NAME}:{DEVICE_INDEX}'
SIZE = (1, 3, 32, 32)

# 1.

full_text = """
<
domain: "slope",
opset_import: [ "" : 1]
>
full (x,shape) => (y1, y2) {
    y1 = Expand(x, shape)
    y2 = Add(x, x)
}
"""
functions = [onnx.parser.parse_function(full_text)]

graph_text = '''
agraph (int64[] shape) => (float[] y1, float[] y2) {    
    one = Constant <value=float[1]{1}>()
    y1, y2 = slope.full(one, shape)
}
'''
graph = onnx.parser.parse_graph(graph_text)
opset_imports = [
            onnx.OperatorSetIdProto(domain="", version=15),
            onnx.OperatorSetIdProto(domain="slope", version=1),
        ]

model = onnx.helper.make_model(
    graph, functions=functions, opset_imports=opset_imports
)
model_text = onnx.printer.to_text(model)
print(model_text)

path = "/tmp/model.onnx"
onnx.save_model(model, path)
sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
input = dict( shape=np.array([2]) )
output_names = ['y1', 'y2']
out = sess.run(output_names, input)
for n, o in zip(output_names, out):
    print(f"{n}\nshape: {o.shape=}\ncontents: {o}\n")

# onnx.checker.check_model(model)



import tensorflow as tf
import tensorflow.mlir.experimental as tme

import slope

7

# print('\n#1')
# x = tf.ones(8, dtype=tf.float32);print(f"{x=}")
# w = tf.constant([[4], [3], [1], [7]], dtype=tf.int32);print(f"{w=}")
# u = tf.constant([9., 10., 11., 12.], dtype=tf.float32);print(f"{u=}")
# y = tf.tensor_scatter_nd_add(x,w,u);print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.tensor_scatter_nd_add)).get_concrete_function(x, w, u,0))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))

'''
module {
  func.func @__inference_tensor_scatter_add_13(%arg0: tensor<8xf32>, %arg1: tensor<4x1xi32>, %arg2: tensor<4xf32>) -> tensor<8xf32> attributes {allow_soft_placement = false} {
    %0 = "stablehlo.scatter"(%arg0, %arg1, %arg2) ({
    ^bb0(%arg3: tensor<f32>, %arg4: tensor<f32>):
      %1 = stablehlo.add %arg3, %arg4 : tensor<f32>
      stablehlo.return %1 : tensor<f32>
    }) {indices_are_sorted = false, scatter_dimension_numbers = #stablehlo.scatter<inserted_window_dims = [0], scatter_dims_to_operand_dims = [0], index_vector_dim = 1>, unique_indices = false} : (tensor<8xf32>, tensor<4x1xi32>, tensor<4xf32>) -> tensor<8xf32>
    return %0 : tensor<8xf32>
  }
}
'''

# print('\n#2')
# x    = tf.constant([[[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
#             [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
#             [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]],
#             [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]]], dtype=tf.float32); print(f"{x=}")
# w = tf.constant([[0], [2]], dtype=tf.int32); print(f"{w=}")
# u = tf.constant([[[5, 5, 5, 5], [6, 6, 6, 6], [7, 7, 7, 7], [8, 8, 8, 8]],
#             [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]]], dtype=tf.float32); print(f"{u=}")

# y = tf.tensor_scatter_nd_update(x,w,u);print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.tensor_scatter_nd_update)).get_concrete_function(x, w, u))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))


# y = tf.tensor_scatter_nd_add(x,w,u);print(f"{y=}")
# f =  tme.convert_function((tf.function(tf.tensor_scatter_nd_add)).get_concrete_function(x, w, u))
# print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))


'''
module {
  func.func @__inference_tensor_scatter_nd_update_11(
    %arg0: tensor<4x4x4xf32>, %arg1: tensor<2x1xi32>, 
    %arg2: tensor<2x4x4xf32>) -> tensor<4x4x4xf32> attributes {allow_soft_placement = false} {
    %0 = "stablehlo.scatter"(%arg0, %arg1, %arg2) ({
    ^bb0(%arg3: tensor<f32>, %arg4: tensor<f32>):
      stablehlo.return %arg4 : tensor<f32>
    }) {
        indices_are_sorted = false,
        scatter_dimension_numbers = #stablehlo.scatter<update_window_dims = [1, 2],
        inserted_window_dims = [0],
        scatter_dims_to_operand_dims = [0],
        index_vector_dim = 1>, unique_indices = false} : (tensor<4x4x4xf32>, tensor<2x1xi32>, tensor<2x4x4xf32>) -> tensor<4x4x4xf32>
    return %0 : tensor<4x4x4xf32>
  }
}
'''

## update
# output  = [[[5, 5, 5, 5], [6, 6, 6, 6], [7, 7, 7, 7], [8, 8, 8, 8]],
#             [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
#             [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]],
#             [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]]]

## add
# array([
# [[ 6.,  7.,  8.,  9.],
#         [11., 12., 13., 14.],
#         [15., 14., 13., 12.],
#         [12., 11., 10.,  9.]],

#        [[ 1.,  2.,  3.,  4.],
#         [ 5.,  6.,  7.,  8.],
#         [ 8.,  7.,  6.,  5.],
#         [ 4.,  3.,  2.,  1.]],

#        [[ 9.,  8.,  7.,  6.],
#         [ 6.,  5.,  4.,  3.],
#         [ 4.,  5.,  6.,  7.],
#         [ 9., 10., 11., 12.]],

#        [[ 8.,  7.,  6.,  5.],
#         [ 4.,  3.,  2.,  1.],
#         [ 1.,  2.,  3.,  4.],
#         [ 5.,  6.,  7.,  8.]]], dtype=float32)>

print('\n#3')
x = tf.zeros((2,2), dtype=tf.float32);print(f"{x=}")
w = tf.constant([[1,0],[0,1]], dtype=tf.int32);print(f"{w=}")
u = tf.constant([1,2], dtype=tf.float32);print(f"{u=}")
y = tf.tensor_scatter_nd_add(x,w,u);print(f"{y=}")
f =  tme.convert_function((tf.function(tf.tensor_scatter_nd_add)).get_concrete_function(x, w, u,0))
print(tme.run_pass_pipeline(f, "tf-lower-to-mlprogram-and-hlo"))
'''
func.func @__inference_tensor_scatter_add_13(%arg0: tensor<2x2xf32>, %arg1: tensor<2x2xi32>, %arg2: tensor<2xf32>) -> tensor<2x2xf32> attributes {allow_soft_placement = false} {
    %0 = "stablehlo.scatter"(%arg0, %arg1, %arg2) ({
    ^bb0(%arg3: tensor<f32>, %arg4: tensor<f32>):
      %1 = stablehlo.add %arg3, %arg4 : tensor<f32>
      stablehlo.return %1 : tensor<f32>
    }) {indices_are_sorted = false, scatter_dimension_numbers = #stablehlo.scatter<inserted_window_dims = [0, 1], scatter_dims_to_operand_dims = [0, 1], index_vector_dim = 1>, unique_indices = false} : (tensor<2x2xf32>, tensor<2x2xi32>, tensor<2xf32>) -> tensor<2x2xf32>
    return %0 : tensor<2x2xf32>
  }
}

'''


__all__ = ["core", "nn", "operators", "procedures"]
__docformat__ = "markdown"
import os
from slope import core
import numpy as np
np.set_printoptions(precision=5, threshold=1000, edgeitems=5, linewidth=120)
SLOPE_BACKEND = os.environ.get("SLOPE_BACKEND", "onnxruntime")
core.set_backend(SLOPE_BACKEND)

def __getattr__(attr):
    if attr in (globals_dict := globals()):
        core.dblog(
            f"Looking slope.{attr} in globals()",
            enable=core.backend.LOG_BACKEND,
        )
        return globals_dict[attr]
    elif attr in vars(core):
        core.dblog(f"Looking slope.{attr} in core", enable=core.backend.LOG_BACKEND)
        return getattr(core, attr)
    elif attr in vars(core.backend.operator_set):
        core.dblog(
            f"Looking slope.{attr} in core.backend.operator_set",
            enable=core.backend.LOG_BACKEND,
        )
        return getattr(core.backend.operator_set, attr)
    elif attr in vars(core.backend.procedure_set):
        core.dblog(
            f"Looking slope.{attr} in core.backend.procedure_set",
            enable=core.backend.LOG_BACKEND,
        )
        return getattr(core.backend.procedure_set, attr)
    elif attr in vars(core.Backend):
        core.dblog(
            f"Looking slope.{attr} in core.backend",
            enable=core.backend.LOG_BACKEND,
        )
        return getattr(core.backend, attr)
    elif attr in core.dtypes.name_dtype_map.keys():
        return core.dtypes.name_dtype_map[attr]
    elif attr in core.dtypes.mlir_dtype_map.keys():
        return core.dtypes.mlir_dtype_map[attr]

    raise NameError(attr)



"""
# run with different backends
SLOPE_BACKEND=iree python examples/nn/cifar10_convnet.py
SLOPE_BACKEND=onnxruntime python examples/nn/cifar10_convnet.py

# print backend code
LOG_JIT=1 SLOPE_BACKEND=iree python examples/nn/cifar10_convnet.py
"""

import slope
import slope.nn as nn

import time
import numpy as np
from tqdm import tqdm

import numpy as np
from lib.datasets.cifar10 import get_cifar10
from lib.models.cv.resnet_cifar import resnet

np.random.seed(12345)


@slope.jit
def train_step(model, batch, optimizer):
    def train_loss_fn(model, batch):
        x, y = batch
        logits, model = model(x, training=True)
        # logits = model(x, training=False)
        loss = logits.cross_entropy(y) / x.size(0)

        return loss, (model, logits)

    (loss, (model, logits)), gmodel = slope.value_and_grad(train_loss_fn, has_aux=True)(model, batch)
    model, optimizer = optimizer(model, gmodel)
    return loss, logits, model, optimizer, gmodel


@slope.jit
def test_step(model, batch):
    x, y = batch
    logits = model(x, training=False)
    loss = logits.cross_entropy(y) / x.size(0)
    return loss, logits


def get_dataloader(images, labels, batch_size, shuffle=False):
    N = images.shape[0]
    perm = np.random.permutation(N) if shuffle else np.arange(N)
    mean = slope.tensor([0.4914, 0.4822, 0.4465], dtype=slope.float32)[None, ..., None, None]
    std = slope.tensor([0.2023, 0.1994, 0.2010], dtype=slope.float32)[None, ..., None, None]

    @slope.jit
    def standardize(x):
        return (x - mean) / std

    def data_iter():
        for i in range(N // batch_size):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            x = standardize(slope.tensor(images[batch_idx], dtype=slope.float32))
            y = slope.tensor(labels[batch_idx], dtype=slope.int32)
            yield x, y

    nbatches = (images.shape[0] // batch_size) + (1 if (images.shape[0] % batch_size != 0) else 0)
    return data_iter, nbatches


if __name__ == "__main__":
    nepochs = 160
    train_batch_size = 100  # TODO: must be multiple of dataset
    test_batch_size = 100
    train_images, train_labels, test_images, test_labels = get_cifar10()

    model = resnet(depth=20)
    optimizer = nn.SGD(model, lr=0.1, momentum=0.9, weight_decay=1e-4)

    train_dataloader, ntrain_batches = get_dataloader(train_images, train_labels, train_batch_size, shuffle=True)
    test_dataloader, ntest_batches = get_dataloader(test_images, test_labels, test_batch_size, shuffle=False)

    print("\nStarting training...")
    best_acc = 0.0
    for epoch in range(nepochs):
        if epoch in [80, 120]:
            optimizer.hp.lr = optimizer.hp.lr * 0.1
            print(f"Optimzer lr now={optimizer.hp.lr.numpy()}")
        total_loss = 0.0
        true_positives = 0.0
        for i, batch in (pbar := tqdm(enumerate(train_dataloader()), total=ntrain_batches, leave=False)):
            # with slope.core.Timing("RUN "): loss, logits, model, optimizer, gmodel = train_step(model, batch, optimizer)
            loss, logits, model, optimizer, gmodel = train_step(model, batch, optimizer)
            loss_numpy = float(loss.numpy())
            total_loss += loss_numpy
            y_hat, y = logits.argmax(-1), batch[1]
            true_positives += float((y_hat == y).cast(slope.float32).sum().numpy())
            train_loss = total_loss / (i + 1)
            train_acc = true_positives / (train_batch_size * (i + 1))
            msg = f"Train epoch: {epoch}, " f"batch_loss: {loss_numpy:.4f}, " f"loss: {train_loss:.4f}, acc: {train_acc:.4f}"
            pbar.set_description(msg)
        mins, secs = divmod(pbar.format_dict["elapsed"], 60)
        print(f"{msg}, Time taken: {int(mins)}:{secs:.2f}")

        total_loss = 0.0
        true_positives = 0.0
        for i, batch in (pbar := tqdm(enumerate(test_dataloader()), total=ntest_batches, leave=False)):
            (
                loss,
                logits,
            ) = test_step(model, batch)
            loss_numpy = float(loss.numpy())
            total_loss += loss_numpy
            y_hat, y = logits.argmax(-1), batch[1]
            true_positives += float((y_hat == y).cast(slope.float32).sum().numpy())
            test_loss = total_loss / (i + 1)
            test_acc = true_positives / (test_batch_size * (i + 1))
            msg = f"Test epoch: {epoch}, " f"batch_loss: {loss_numpy:.4f}, " f"loss: {test_loss:.4f}, acc: {test_acc:.4f}"
            pbar.set_description(msg)
        mins, secs = divmod(pbar.format_dict["elapsed"], 60)
        print(f"{msg}, Time taken: {int(mins)}:{secs:.2f}")
        if test_acc > best_acc:
            print(f"  Best accuracy {test_acc:.4f} at epoch {epoch}\n")
            best_acc = test_acc



# A set of code samples showing different usage of the ONNX Runtime Python API
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import onnxruntime
import onnx
# from onnx import parser, printer

MODEL_FILE = '.model.onnx'
DEVICE_NAME = 'cpu'
DEVICE_INDEX = 0
DEVICE=f'{DEVICE_NAME}:{DEVICE_INDEX}'
SIZE = (1, 3, 32, 32)


model_text = """
<
   ir_version: 9,
   opset_import: ["" : 15]
>
agraph (int64[] x) => (int64[] y) {
   c = Constant <value = true> ()
   y= And(c)
}

"""

model = onnx.parser.parse_model(model_text)
# onnx.checker.check_model(model)
model_text = onnx.printer.to_text(model)
print(model_text)

path = "/tmp/model.onnx"
onnx.save_model(model, path)
sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
input = dict(x=np.array([True]))
output_names = ['y']
out = sess.run(output_names, input)
for o in out:
    print(f"{o.shape=}\n{o=}")


model_text = """
<
   ir_version: 9,
   opset_import: ["" : 15]
>
agraph (int64[] l) => (int64[] y) {
   s=Constant <value_int = 0> ()
   d=Constant <value_int = 1> ()
   y= Range(s, l, d)
}

"""

model = onnx.parser.parse_model(model_text)
# onnx.checker.check_model(model)
model_text = onnx.printer.to_text(model)
print(model_text)

path = "/tmp/model.onnx"
onnx.save_model(model, path)
sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
input = dict(l=np.array([5]))
output_names = ['y']
out = sess.run(output_names, input)
for o in out:
    print(f"{o.shape=}\n{o=}")

# model_text = """
# <
#    ir_version: 9,
#    opset_import: ["" : 15, "slope" : 1]
# >
# agraph (int64[] shape) => (float[] y, float[] z,) {
#    one = Constant <value = float[1] {1}> ()
#    y, z = slope.full (one, shape)
# }
# <
#   domain: "slope",
#   opset_import: ["" : 1]
# >
# full (x, shape) => (y, z)
# {
#    y = Expand (x, shape)
#    z = x
# }
# """

# model = onnx.parser.parse_model(model_text)
# # onnx.checker.check_model(model)
# model_text = onnx.printer.to_text(model)
# print(model_text)

# path = "/tmp/model.onnx"
# onnx.save_model(model, path)
# sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
# input = dict()
# output_names = ['y']
# out = sess.run(output_names, input)
# for o in out:
#     print(f"{o.shape=}\n{o=}")
    

# model_text = """
# <ir_version: 9, opset_import: ["" : 17, "slope" : 1]>
# model () => (float[] y) {
#    y = Constant <value = float[1] {1.0} > ()
# }
# """

# model = onnx.parser.parse_model(model_text)
# # onnx.checker.check_model(model)
# model_text = onnx.printer.to_text(model)
# print(model_text)

# path = "/tmp/model.onnx"
# onnx.save_model(model, path)
# sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
# input = dict()
# output_names = ['y']
# out = sess.run(output_names, input)
# for o in out:
#     print(f"{o.shape=}\n{o=}")
    
# model_text = """
# <ir_version: 9, opset_import: ["" : 17, "slope" : 1]>
# model () => (float[] y) {
#    y = Constant <value = float[1] {1}> ()
# }
# """

# model = onnx.parser.parse_model(model_text)
# model_text = onnx.printer.to_text(model)
# print(model_text)

# path = "/tmp/model.onnx"
# onnx.save_model(model, path)
# sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
# input = dict()
# output_names = ['y']
# out = sess.run(output_names, input)
# for o in out:
#     print(f"{o.shape=}\n{o=}")

# model_text = """
# <ir_version: 9, opset_import: ["" : 17, "slope" : 1]>
# model (int64[] shape) => (float[] y) {
#    y = slope.ones (shape)
# }

# <domain: "slope", opset_import: ["" : 17]>
# full (x, shape) => (y)
# {
#    y = Expand (x, shape)
# }

# <domain: "slope", opset_import: ["" : 17, "slope" : 1]>
# ones (shape) => (y)
# {
#    one = Constant <value = float[1] {1}> ()
# y = slope.full(one, shape)
# }
# """
# model = onnx.parser.parse_model(model_text)
# model_text = onnx.printer.to_text(model)
# print(model_text)

# path = "/tmp/model.onnx"
# onnx.save_model(model, path)
# sess= onnxruntime.InferenceSession(path,  providers=['CPUExecutionProvider'])
# input = dict( shape=np.array([2]) )
# output_names = ['y']
# out = sess.run(output_names, input)
# for o in out:
#     print(f"{o.shape=}\n{o=}")

# onnx.checker.checkm _model(model)



import slope
import slope.core
from slope.core import (
    Backend,
    Operator,
    Instruction,
    Program,
    MetaOperator,
    OperatorSet,
    ProcedureSet,
    Tensor,
    TensorBuffer,
    SymbolicTensor,
    list_map,
    dtypes,
    devices,
    DType,
    Device,
    CodegenOutput,
)

import math
import numpy as np
from typing import (
    List,
    Any,
)
import iree.compiler
import iree.runtime
import os

from slope.operators import operator_set
from slope.procedures import procedure_set
import tempfile
import pygments
from pygments.lexers.jvm import JavaLexer
from pygments.formatters import Terminal256Formatter


def annotate_shape(symval):
    xdtype = symval.dtype.mlir
    if len(symval.shape) > 0:
        xshape = f"{'x'.join((repr(i) for i in symval.shape))}"
        return f"tensor<{xshape}x{xdtype}>"
    else:
        return f"tensor<{xdtype}>"


def annotate_sig(in_symvals, out_symvals):
    if isinstance(in_symvals, SymbolicTensor):
        in_symvals = (in_symvals,)
    if isinstance(out_symvals, SymbolicTensor):
        out_symvals = (out_symvals,)
    in_sig = f"({','.join(annotate_shape(t) for t in in_symvals)})"
    out_sig = f"({','.join(annotate_shape(t) for t in out_symvals)})"
    sig = f"{in_sig} -> {out_sig}"
    return sig


class IREEBackend(Backend):
    dtype_for_indices = dtypes.int32
    dtype_map = {
        dtypes.float32: np.dtypes.Float32DType(),
        dtypes.uint8: np.dtypes.UInt8DType(),
        dtypes.int8: np.dtypes.Int8DType(),
        dtypes.bool: np.dtypes.BoolDType(),
        dtypes.int32: np.dtypes.Int32DType(),
        dtypes.int64: np.dtypes.Int64DType(),
        dtypes.uint64: np.dtypes.UInt64DType(),
        dtypes.float16: np.dtypes.Float16DType(),
    }
    device_map = {
        devices.cpu: "local-task",
        devices.cuda: "cuda",
        devices.metal: "metal",
    }

    target_map = {
        devices.cpu: "llvm-cpu",
        devices.cuda: "cuda",
        devices.metal: "metal",
    }

    dtype_map_inv = {v: k for k, v in dtype_map.items()}
    device_map_inv = {v: k for k, v in device_map.items()}

    def from_numpy(self, val, dtype=None, device=None):
        dtype = dtype or self.DEFAULT_DTYPE
        device = device or self.DEFAULT_DEVICE
        np_val = np.array(val, dtype=dtype.numpy)
        iree_device = iree.runtime.get_device(self.device_map[device])
        val = iree.runtime.asdevicearray(iree_device, np_val)
        return Tensor(TensorBuffer(val))

    def numpy_of(self, tensor: Tensor, memmap=False):
        if not memmap:
            return tensor.buf.val.to_host()
        with tempfile.NamedTemporaryFile() as arr_file:
            arr = np.memmap(arr_file.name, dtype=tensor.dtype.numpy, shape=tensor.shape, mode="w+")
            arr[:] = tensor.buf.val.to_host()
            return arr

    def device_of(self, tensor: Tensor):
        return self.device_map_inv[str(tensor.buf.val._device)]

    def shape_of(self, tensor: Tensor):
        return tuple(tensor.buf.val.shape)

    def dtype_of(self, tensor: Tensor):
        return self.dtype_map_inv[tensor.buf.val.dtype]

    def codegen(self, program: Program, args, *, fn_name: str = "main", fn_defs=dict()) -> List[Any]:
        if fn_name == "main":
            assert not hasattr(self, "fn_count")
            self.fn_count = 0

        def indent(code):
            spaces = " " * (len(code) - len(code.lstrip()))
            spaces += " " * 4
            return "\n".join([spaces + line for line in code.strip().split("\n")])

        # codegen is recursive if jit-of-jit happens
        body_code_lines = []

        for instruction in program.instructions:
            if len(instruction.out_binders) == 0:  # skip codegen for function returns nothing
                continue
            in_vals = list_map(lambda x: program.env[x], instruction.inputs)
            out_vals = list_map(lambda z: program.env[z], instruction.out_binders)
            if isinstance(instruction.op, MetaOperator):
                impl_code, fn_defs = self.impls[instruction.op](args, instruction, fn_defs, in_vals, out_vals)
            else:
                if instruction.op in self.impls.keys():
                    impl_code = self.impls[instruction.op](*in_vals, *out_vals, **instruction.params)
                else:
                    # No impl is defined, fallback to procedure
                    impl_code, fn_defs = self.codegen_impl_as_procedure(args, instruction, fn_defs, in_vals, out_vals)

            for impl_code_line in impl_code.split("\n"):  # handle multi-line code
                body_code_lines += [indent(impl_code_line)]

        in_binders = list_map(lambda x: program.env[x], program.in_binders)
        fn_args_str = ", ".join([f"%{i.name}: {annotate_shape(i.symval)}" for i in in_binders])

        outs = list_map(lambda x: program.env[x], program.outs)
        out_str = ", ".join([f"%{o.name}" for o in outs])
        out_type_str = ", ".join([f"{annotate_shape(o.symval)}" for o in outs])

        head_code_line = [f"func.func @{fn_name} ({fn_args_str}) -> ({out_type_str})"]
        tail_code_line = [indent(f'"func.return"({out_str}): ({out_type_str}) -> ()')]
        model_code_lines = head_code_line + ["{"] + body_code_lines + tail_code_line + ["}"]

        code_lines = model_code_lines
        if fn_name == "main":
            functions_code_lines = []
            for fn_def_code_lines in fn_defs.values():
                functions_code_lines += fn_def_code_lines
            code_lines += functions_code_lines

        if slope.LOG_JIT:
            formatted_code = pygments.highlight("\n".join(code_lines), JavaLexer(), Terminal256Formatter())
            slope.core.dblog(f"\n---- {program.name} codegen:\n\n" + formatted_code + "\n\n===============\n")

        if fn_name == "main":
            del self.fn_count
        assert len(outs) == len(program.outs)
        return CodegenOutput(
            code_lines=code_lines,
            fn_defs=fn_defs,
            in_binders=in_binders,
            outs=outs,
        )

    def codegen_impl_as_procedure(self, args, instruction: Instruction, fn_defs, in_vals, out_vals):
        op: Operator = instruction.op
        op_name = {v: k for (k, v) in vars(self.operator_set).items()}[op]
        op_procedure = getattr(self.procedure_set, op_name)
        symvals_in = tuple(inp.symval for inp in instruction.inputs)
        params = instruction.params
        op_program, consts, _ = slope.core.make_program(
            op_procedure,
            *symvals_in,
            static_args=tuple(params.items()),
            name=op.name,
        )
        name = slope.core.jit.get_jit_name(tuple(symvals_in), params)
        if name not in fn_defs.keys():
            # if name == "jit_shape__lp_50_cm_3_cm_32_cm_32_rp__dtype_float32_shape__lp_16_cm_3_cm_3_cm_3_rp__dtype_float32_groups_1_stride__lp_1_cm_1_rp__dilation__lp_1_cm_1_rp__padding__lp_1_cm_1_cm_1_cm_1_rp__":
            #     breakpoint()
            op_codegen_output: CodegenOutput = self.codegen(
                op_program,
                args,
                fn_name=name,
                fn_defs=fn_defs,
            )
            fn_defs = {**fn_defs, **op_codegen_output.fn_defs}
            fn_defs[name] = op_codegen_output.code_lines
        in_names = ", ".join(f"%{i.name}" for i in in_vals)
        out_names = ", ".join(f"%{o.name}" for o in out_vals)
        sig = annotate_sig(tuple(i.symval for i in in_vals), out_vals[0].symval)
        impl_code = f"{out_names} = func.call @{name}({in_names}) : {sig}"
        return impl_code, fn_defs

    def compile(self, codegen_output):
        code_lines = codegen_output.code_lines
        code = "\n".join(code_lines)
        instance = iree.runtime.VmInstance()
        device = codegen_output.outs[0].device
        iree_device = iree.runtime.get_device(self.device_map[device])
        hal_module = iree.runtime.create_hal_module(instance, iree_device)
        binary = iree.compiler.compile_str(
            code,
            target_backends=(self.target_map[device],),
            optimize=True,
            extra_args=[
                "--iree-vm-bytecode-module-optimize",
                # "--iree-opt-const-eval",
                # "--iree-opt-const-expr-hoisting",
                # "--iree-opt-data-tiling",
                # "--iree-opt-numeric-precision-reduction"
            ],
        )
        m = iree.runtime.VmModule.from_flatbuffer(instance, binary)
        context = iree.runtime.VmContext(instance, modules=[hal_module, m])
        f = m.lookup_function("main")
        # finv = iree.runtime.FunctionInvoker(context, iree_device, f)
        finv = iree.runtime.FunctionInvoker(context, iree_device, f, tracer=None)

        return finv, code

    def export(self, jit_output, output_path, export_params, input_names, output_names, **kwargs):
        # iree.compiler.core.DEFAULT_TESTING_BACKENDS
        target_backends = kwargs.get("target_backends", "llvm-cpu")
        if isinstance(target_backends, str):
            target_backends = (target_backends,)
        os.makedirs(output_path, exist_ok=True)

        code_lines = jit_output.codegen_output.code_lines[:]
        in_binders = jit_output.codegen_output.in_binders
        outs = jit_output.codegen_output.outs
        num_consts = jit_output.program.num_consts
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            for i in range(num_consts):
                const_pattern = f"%{in_binders[i].name}: {annotate_shape(in_binders[i].symval)}"
                code_lines[0] = code_lines[0].replace(f"{const_pattern}, ", "")
                code_lines[0] = code_lines[0].replace(f"{const_pattern})", ")")
            f.write("\n".join(code_lines[:2]) + "\n")
            for i in range(num_consts):
                f.write(
                    f"    %{in_binders[i].name} = stablehlo.constant "
                    f"""dense<"0x{in_binders[i].symval.numpy(memmap=True).tobytes().hex()}">"""
                    f": {annotate_shape(in_binders[i].symval)}\n"
                )
            f.writelines("\n".join(code_lines[2:]))
            f.flush()
            iree.compiler.compile_file(f.name, target_backends=target_backends, output_file=os.path.join(output_path, "model.vmfb"))

        input_arg_names = [ib.name for ib in in_binders[num_consts:]]
        input_arg_names_str = ", ".join(input_arg_names)
        outs_names = [out.name for out in outs]

        test_input_code = ""
        for i in range(num_consts, len(in_binders)):
            input_name = in_binders[i].name
            input_shape = in_binders[i].symval.shape
            dtype = in_binders[i].symval.dtype
            input_dtype = ("np." + dtype.numpy.__name__) if dtype is not dtypes.bool else "bool"
            test_input_code += f"""    {input_name} = np.ones({input_shape}, dtype={input_dtype})\n"""

        module_path = os.path.join(output_path, "__init__.py")
        module_code = f"""#!/usr/bin/env python
import os
import numpy as np
import iree.runtime

root_path = os.path.dirname(__file__)
model_path = os.path.join(root_path, "model.vmfb")
input_arg_names = {input_arg_names}
out_names = {outs_names}

instance = iree.runtime.VmInstance()
iree_device = iree.runtime.get_device("local-task")
hal_module = iree.runtime.create_hal_module(instance, iree_device)
with open(model_path, 'r+b') as f:
    m = iree.runtime.VmModule.from_flatbuffer(instance, f.read())
context = iree.runtime.VmContext(instance, modules=[hal_module, m])
f = m.lookup_function("main")
finv = iree.runtime.FunctionInvoker(context, iree_device, f, tracer=None)
if __name__ == "__main__":
{test_input_code}
    print("inputs:")
    for inp_name, inp in zip(input_arg_names, ({input_arg_names_str})):
        print(f"{{inp_name}} {{inp.shape}}{{inp.dtype}}\\n{{inp}}\\n")
    outs = finv({input_arg_names_str})
    if not isinstance(outs, tuple):
        outs = (outs,)
    print("outputs:")
    for out_name, out in zip(out_names, outs):
        print(f"{{out_name}} {{out.shape}}{{out.dtype}}\\n{{out.to_host()}}\\n")
"""
        with open(module_path, "w") as f:
            f.write(module_code)
            slope.dblog(module_code, enable=slope.LOG_JIT)


backend = IREEBackend(operator_set, procedure_set)


@backend.set_impl(backend.operator_set.jit_op)
def jit_op_impl(self, args, instruction, fn_defs, in_vals, out_vals):
    jit_program = instruction.params["program"]
    jit_name = f"{jit_program.name}"
    if jit_name not in fn_defs.keys():
        jit_codegen_output = self.codegen(
            jit_program,
            args,
            fn_name=jit_name,
            fn_defs=fn_defs,
        )
        if (
            jit_name
            == "@jit_shape__lp_50_cm_3_cm_32_cm_32_rp__dtype_float32_shape__lp_16_cm_3_cm_3_cm_3_rp__dtype_float32_groups_1_stride__lp_1_cm_1_rp__dilation__lp_1_cm_1_rp__padding__lp_1_cm_1_cm_1_cm_1_rp__@jit_shape__lp_50_cm_3_cm_32_cm_32_rp__dtype_float32_shape__lp_16_cm_3_cm_3_cm_3_rp__dtype_float32_groups_1_stride__lp_1_cm_1_rp__dilation__lp_1_cm_1_rp__padding__lp_1_cm_1_cm_1_cm_1_rp__"
        ):
            breakpoint()
        fn_defs = {**fn_defs, **jit_codegen_output.fn_defs}
        fn_defs[jit_name] = jit_codegen_output.code_lines

    args_str = ", ".join(i.name for i in in_vals)
    sig = annotate_sig(tuple(i.symval for i in in_vals), out_vals[0].symval)
    impl_code = f"{', '.join(o.name for o in out_vals)} = func.call @{jit_name}({args_str}) {sig}"
    return impl_code, fn_defs


@backend.set_impl(backend.operator_set.cast)
def cast_impl(self, x, y, *, dtype):
    return f'%{y.name} = "stablehlo.convert"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.stop_gradient)
def stop_gradient_impl(self, x, y):
    return f'%{y.name} = "stablehlo.optimization_barrier"(%{x.name}): {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.sqrt)
def sqrt_impl(self, x, y):
    return f'%{y.name} = "stablehlo.sqrt"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.exp)
def exp_impl(self, x, y):
    return f'%{y.name} = "stablehlo.exponential"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.log)
def log_impl(self, x, y):
    return f'%{y.name} = "stablehlo.log"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.sin)
def sin_impl(self, x, y):
    return f'%{y.name} = "stablehlo.sine"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.invert)
def invert_impl(self, x, y):
    return f'%{y.name} = "stablehlo.not"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.add)
def add_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.add"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


@backend.set_impl(backend.operator_set.sub)
def sub_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.subtract"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


@backend.set_impl(backend.operator_set.mul)
def mul_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.multiply"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


@backend.set_impl(backend.operator_set.div)
def div_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.divide"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


@backend.set_impl(backend.operator_set.pow)
def pow_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.power"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


def get_compare_type(dtype):
    return "FLOAT" if dtypes.is_float(dtype) else "SIGNED" if dtypes.is_int(dtype) and dtype is not dtypes.uint8 else "UNSIGNED"


@backend.set_impl(backend.operator_set.equal)
def equal_impl(self, x, w, y):
    return f"""%{y.name} = "stablehlo.compare"(%{x.name}, %{w.name}) {{
  comparison_direction = #stablehlo<comparison_direction EQ>,
  compare_type = #stablehlo<comparison_type {get_compare_type(x.symval.dtype)}>
}}  : {annotate_sig((x.symval, w.symval), y.symval)}
"""


@backend.set_impl(backend.operator_set.less)
def less_impl(self, x, w, y):
    return f"""%{y.name} = "stablehlo.compare"(%{x.name}, %{w.name}) {{
  comparison_direction = #stablehlo<comparison_direction LT>,
  compare_type = #stablehlo<comparison_type {get_compare_type(x.symval.dtype)}>
}}  : {annotate_sig((x.symval, w.symval), y.symval)}
"""


@backend.set_impl(backend.operator_set.greater)
def greater_impl(self, x, w, y):
    return f"""%{y.name} = "stablehlo.compare"(%{x.name}, %{w.name}) {{
  comparison_direction = #stablehlo<comparison_direction GT>,
  compare_type = #stablehlo<comparison_type {get_compare_type(x.symval.dtype)}>
}}  : {annotate_sig((x.symval, w.symval), y.symval)}
"""


@backend.set_impl(backend.operator_set.maximum)
def maximum_impl(self, x, w, y):
    return f'%{y.name} = "stablehlo.maximum"(%{x.name}, %{w.name}) : {annotate_sig((x.symval, w.symval), y.symval)}'


@backend.set_impl(backend.operator_set.matmul)
def matmul_impl(self, x, w, y):
    x_bdims = [] if (w.symval.ndim <= 2 or x.symval.ndim == 1) else list(range(x.symval.ndim - (2 if x.symval.ndim > 2 else 1)))
    w_bdims = [] if (w.symval.ndim == 1 or x.symval.ndim <= 2) else list(range(w.symval.ndim - 2))
    x_cdim = 0 if x.symval.ndim == 1 else x.symval.ndim - 1
    w_cdim = 0 if w.symval.ndim == 1 else w.symval.ndim - 2
    return f"""%{y.name} = "stablehlo.dot_general"(%{x.name}, %{w.name}) {{
  dot_dimension_numbers = #stablehlo.dot<
    lhs_batching_dimensions = {x_bdims},
    rhs_batching_dimensions = {w_bdims},
    lhs_contracting_dimensions = [{x_cdim}],
    rhs_contracting_dimensions = [{w_cdim}]
  >,
  precision_config = [#stablehlo<precision DEFAULT>, #stablehlo<precision DEFAULT>]
    }}  : {annotate_sig((x.symval, w.symval), y.symval)}
"""


@backend.set_impl(backend.operator_set.sum)
def sum_impl(self, x, y, *, dim, keepdim):
    zero = "0." if dtypes.is_float(y.symval.dtype) else "0"
    y_init_type = SymbolicTensor((), y.symval.dtype, y.symval.device)
    y_mlir_type = annotate_shape(y_init_type)
    y_out_type = (
        y.symval
        if not keepdim
        else SymbolicTensor(
            tuple(d for i, d in enumerate(y.symval.shape) if i not in dim),
            y.symval.dtype,
            y.symval.device,
        )
    )
    return f"""
%{y.name}_init = stablehlo.constant dense<{zero}> : {annotate_shape(y_init_type)}
%{y.name}{'_' if keepdim else ''} = "stablehlo.reduce"(%{x.name}, %{y.name}_init) ({{
  ^bb0(%arg0: {y_mlir_type}, %arg1: {y_mlir_type}):
    %0 = "stablehlo.add"(%arg0, %arg1) : {annotate_sig((y_init_type, y_init_type), y_init_type)}
    "stablehlo.return"(%0) : ({y_mlir_type}) -> ()
}}) {{
  dimensions = dense<{repr(list(dim))}> : tensor<{len(dim)}xi64>
}} : {annotate_sig((x.symval, y_init_type), y_out_type)}
{f'%{y.name} = "stablehlo.reshape"(%{y.name}_) : {annotate_sig((y_out_type,), y.symval)}' if keepdim else ''}"""


@backend.set_impl(backend.operator_set.max)
def max_impl(self, x, y, *, dim, keepdim):
    min_val = {
        dtypes.float32: "1.E-38",
        dtypes.int8: "-128",
        dtypes.int32: "-65536",
    }[x.symval.dtype]
    y_init_type = SymbolicTensor((), y.symval.dtype, y.symval.device)
    y_mlir_type = annotate_shape(y_init_type)
    y_out_type = (
        y.symval
        if not keepdim
        else SymbolicTensor(
            tuple(d for i, d in enumerate(y.symval.shape) if i not in dim),
            y.symval.dtype,
            y.symval.device,
        )
    )
    return f"""
%{y.name}_init = stablehlo.constant dense<{min_val}> : {annotate_shape(y_init_type)}
%{y.name}{'_' if keepdim else ''} = "stablehlo.reduce"(%{x.name}, %{y.name}_init) ({{
  ^bb0(%arg0: {y_mlir_type}, %arg1: {y_mlir_type}):
    %0 = "stablehlo.maximum"(%arg0, %arg1) : {annotate_sig((y_init_type, y_init_type), y_init_type)}
    "stablehlo.return"(%0) : ({y_mlir_type}) -> ()
}}) {{
  dimensions = dense<{repr(list(dim))}> : tensor<{len(dim)}xi64>
}} : {annotate_sig((x.symval, y_init_type), y_out_type)}
{f'%{y.name} = "stablehlo.reshape"(%{y.name}_) : {annotate_sig((y_out_type,), y.symval)}' if keepdim else ''}
"""


@backend.set_impl(backend.operator_set.arange)
def arange_impl(self, y, *, start, stop, stride, dtype, device):
    if stride == 1 and start == 0:
        return f"""%{y.name} = "stablehlo.iota"() {{iota_dimension = 0 : i64}} : {annotate_sig((), y.symval)}"""
    one_symval = y.symval.like(shape=(1,))
    return f"""
%{y.name}_scale_ = stablehlo.constant dense<{stride}> : {annotate_shape(one_symval)}
%{y.name}_scale = "stablehlo.broadcast_in_dim"(%{y.name}_scale_) {{
        broadcast_dimensions = dense<{repr(list(range(y.symval.ndim)))}>: tensor<{y.symval.ndim}xi64>
        }} : {annotate_sig(( one_symval,), y.symval)}
%{y.name}_shift_ = stablehlo.constant dense<{start}> : {annotate_shape(y.symval.like(shape=(1,)))}
%{y.name}_shift = "stablehlo.broadcast_in_dim"(%{y.name}_shift_) {{
        broadcast_dimensions = dense<{repr(list(range(y.symval.ndim)))}>: tensor<{y.symval.ndim}xi64>
        }} : {annotate_sig(( one_symval,), y.symval)}
%{y.name}__ = "stablehlo.iota"() {{iota_dimension = 0 : i64}} : {annotate_sig((), y.symval)}
%{y.name}_ = "stablehlo.multiply"(%{y.name}__, %{y.name}_scale) : {annotate_sig((y.symval, y.symval), y.symval)}
%{y.name} = "stablehlo.add"(%{y.name}_, %{y.name}_shift) : {annotate_sig((y.symval, y.symval), y.symval)}
"""


@backend.set_impl(backend.operator_set.full)
def full_impl(self, y, *, shape, fill_value, dtype, device):
    fill_value = float(fill_value) if "f" in dtype.mlir else int(fill_value)
    fill_value = repr(fill_value)
    fill_value = fill_value.replace("e", "E") if "." in fill_value else fill_value.replace("e", ".E")
    return (
        f'%{y.name} = "stablehlo.constant"() {{ value = dense<{fill_value}> : {annotate_shape(y.symval)} }} : {annotate_sig((), y.symval)}'
    )


@backend.set_impl(backend.operator_set.random_uniform)
def random_uniform_impl(self, y, *, shape, dtype, device):
    zero = "0." if dtypes.is_float(y.symval.dtype) else "0"
    one = "1." if dtypes.is_float(y.symval.dtype) else "1"
    a_type = b_type = y.symval.like(shape=())
    is_scalar = shape == ()
    shape_val = f'dense<{repr(list(shape)) if not is_scalar else "[1]"}'
    shape_type = y.symval.like(shape=(1,) if is_scalar else (len(shape),), dtype=dtypes.int64)
    y_out_type = y.symval if not is_scalar else y.symval.like(shape=(1,))
    return f"""%{y.name}_a = stablehlo.constant dense<{zero}> : {annotate_shape(a_type)}
%{y.name}_b = stablehlo.constant dense<{one}> : {annotate_shape(b_type)}
%{y.name}_shape = stablehlo.constant {shape_val}> : {annotate_shape(shape_type)}
%{y.name}{'_' if is_scalar else ''} = "stablehlo.rng"(%{y.name}_a, %{y.name}_b,%{y.name}_shape) {{
        rng_distribution = #stablehlo<rng_distribution UNIFORM>
        }} : {annotate_sig((a_type, b_type, shape_type), y_out_type)}
{f'%{y.name} = "stablehlo.reshape"(%{y.name}_) : {annotate_sig((y_out_type,), y.symval)}' if is_scalar else ''}"""


@backend.set_impl(backend.operator_set.random_normal)
def random_normal_impl(self, y, *, shape, dtype, device):
    zero = "0." if dtypes.is_float(y.symval.dtype) else "0"
    one = "1." if dtypes.is_float(y.symval.dtype) else "1"
    a_type = b_type = SymbolicTensor((), dtype, device)
    is_scalar = shape == ()
    shape_val = f'dense<{repr(list(shape)) if not is_scalar else "[1]"}'
    shape_type = SymbolicTensor((1,) if is_scalar else (len(shape),), slope.dtypes.int64, device)
    y_out_type = y.symval if not is_scalar else SymbolicTensor((1,), y.symval.dtype, y.symval.device)
    return f"""%{y.name}_a = stablehlo.constant dense<{zero}> : {annotate_shape(a_type)}
%{y.name}_b = stablehlo.constant dense<{one}> : {annotate_shape(b_type)}
%{y.name}_shape = stablehlo.constant {shape_val}> : {annotate_shape(shape_type)}
%{y.name}{'_' if is_scalar else ''} = "stablehlo.rng"(%{y.name}_a, %{y.name}_b,%{y.name}_shape) {{
        rng_distribution = #stablehlo<rng_distribution NORMAL>}} : {annotate_sig((a_type, b_type, shape_type), y_out_type)}
{f'%{y.name} = "stablehlo.reshape"(%{y.name}_) : {annotate_sig((y_out_type,), y.symval)}' if is_scalar else ''}"""


# @backend.set_impl(backend.operator_set.rng_bits)
# def rng_bits_impl(self, x, y, *, shape, dtype, device):
#     return f"""%{y.name}_, %{y.name} =  "stablehlo.rng_bit_generator"(%{x.name}) {{
#   rng_algorithm = #stablehlo<rng_algorithm THREE_FRY>
# }} : ({annotate_shape(x.symval)}) -> ({annotate_shape(x.symval)}, {annotate_shape(y.symval)})"""


@backend.set_impl(backend.operator_set.expand)
def expand_impl(self, x, y, *, shape):
    return f"""%{y.name} = "stablehlo.broadcast_in_dim"(%{x.name}) {{
        broadcast_dimensions = dense<{repr(list(range(len(shape))))}>: tensor<{len(shape)}xi64>
        }} : {annotate_sig(( x.symval,), y.symval)}
"""


@backend.set_impl(backend.operator_set.reshape)
def reshape_impl(self, x, y, *, shape):
    return f'%{y.name} = "stablehlo.reshape"(%{x.name}) : {annotate_sig((x.symval,), y.symval)}'


@backend.set_impl(backend.operator_set.pad)
def pad_impl(self, x, y, *, padding, mode, value):
    value = float(value) if "f" in x.symval.dtype.mlir else int(value)
    value_type = SymbolicTensor((), x.symval.dtype, x.symval.device)
    lo = padding[0::2][::-1]
    hi = padding[1::2][::-1]
    return f"""%{y.name}_value = stablehlo.constant dense<{value}> : {annotate_shape(value_type)}
%{y.name} = "stablehlo.pad"(%{x.name}, %{y.name}_value) {{
  edge_padding_low = dense<{repr(list(lo))}> : tensor<{len(lo)}xi64>,
  edge_padding_high = dense<{repr(list(hi))}> : tensor<{len(hi)}xi64>,
  interior_padding = dense<{repr([0]*len(lo))}> : tensor<{len(lo)}xi64>
}} : {annotate_sig((x.symval, value_type), y.symval)}
"""


@backend.set_impl(backend.operator_set.slice)
def slice_impl(self, x, y, *, starts, limits, strides):
    return f"""%{y.name} = "stablehlo.slice"(%{x.name}) {{
  start_indices = dense<{repr(list(starts))}> : tensor<{len(starts)}xi64>,
  limit_indices = dense<{repr(list(limits))}> : tensor<{len(limits)}xi64>,
  strides = dense<{repr(list(strides))}> : tensor<{len(strides)}xi64>
}} : {annotate_sig((x.symval,), y.symval)}
"""


@backend.set_impl(backend.operator_set.cat)
def cat_impl(self, *xs, dim):
    xs, y = xs[:-1], xs[-1]
    return f"""%{y.name} = "stablehlo.concatenate"({', '.join([f'%{x.name}' for x in xs])}) {{
 dimension = {dim} : i64
}} : {annotate_sig(([x.symval for x in xs]), y.symval)}"""


@backend.set_impl(backend.operator_set.permute)
def permute_impl(self, x, y, *, perm):
    return f"""%{y.name} = "stablehlo.transpose"(%{x.name}) {{
  permutation = dense<{repr(list(perm))}> : tensor<{len(perm)}xi64>
}} : {annotate_sig((x.symval,), y.symval)}"""


@backend.set_impl(backend.operator_set.flip)
def flip_impl(self, x, y, *, dim):
    return f"""%{y.name} = "stablehlo.reverse"(%{x.name}) {{
  dimensions = dense<{repr(list(dim))}> : tensor<{len(dim)}xi64>
}}  : {annotate_sig((x.symval,), y.symval)}
"""


@backend.set_impl(backend.operator_set.conv)
def conv_impl(self, x, w, y, *, groups, stride, dilation, padding):
    padding = [[s, e] for s, e in zip(list(padding[0::2]), list(padding[1::2]))]
    D = len(x.symval.shape[2:])
    nD = repr(list(range(D)))[1:-1]
    xdims = f"[b, f, {nD}]"
    wdims = f"[o, i, {nD}]"
    return f"""%{y.name} = "stablehlo.convolution"(%{x.name}, %{w.name}) {{
  window_strides = dense<{list(stride)}> : tensor<{len(stride)}xi64>,
  padding = dense<{padding}> : tensor<{D}x{D}xi64>,
  lhs_dilation = dense<1> : tensor<{D}xi64>,
  rhs_dilation = dense<{list(dilation)}> : tensor<{D}xi64>,
  window_reversal = dense<false> : tensor<{D}xi1>,
  dimension_numbers = #stablehlo.conv<{xdims}x{wdims}->{xdims}>,
  feature_group_count = {groups} : i64,
  batch_group_count = 1 : i64,
  precision_config = [#stablehlo<precision DEFAULT>, #stablehlo<precision DEFAULT>]
}}  : {annotate_sig((x.symval, w.symval), y.symval)}
"""


# @backend.set_impl(backend.operator_set.where)
# def where_impl(self, x, w, u, y):
# return f"""%{y.name} = "stablehlo.select"(%{x.name}, %{w.name}, %{u.name}) : {annotate_sig((x.symval,w.symval,u.symval), y.symval)}"""


@backend.set_impl(backend.operator_set.gather_nd)
def gather_nd_impl(self, x, w, y, *, batch_dims):
    operand_shape = list(x.symval.shape)
    indices_shape = list(w.symval.shape)
    r = x.symval.ndim
    q = w.symval.ndim
    b = batch_dims
    offset_dims = list(range(1, q))
    index_vector_dim = q - 1
    y_reshape = None
    w_arange = None
    if b > 0:
        w_arange = w.symval

    if indices_shape[-1] == r:
        slice_sizes = [1] * r
        start_index_map = [i for i in range(q)]
        if b != 0:
            start_index_map = start_index_map[b:] + start_index_map[:b]
        collapsed_slice_dims = [
            i for i in range(1 - b, len(slice_sizes) - b) if slice_sizes[i] == 1 and operand_shape[i] == indices_shape[i]
        ]

        y_reshape = y.symval.like(shape=y.symval.shape + (1,))
    elif indices_shape[-1] < r:
        slice_sizes = [*[1] * (r - 1), *operand_shape[-1:]]
        start_index_map = [i for i, s in enumerate(slice_sizes) if s == 1 and i < q]

        collapsed_slice_dims = []
        for i in range(len(slice_sizes)):
            if slice_sizes[i] == 1 and len(offset_dims) + len(collapsed_slice_dims) != r:
                collapsed_slice_dims += [i]

        if len(collapsed_slice_dims) != len(start_index_map):
            y_reshape = y.symval.like(shape=y.symval.shape + (1,))

    else:
        raise ValueError
    w_fixed = (
        w.symval
        if w_arange is None
        else w.symval.like(shape=tuple(d + b if i == b else d for i, d in enumerate(w.symval.shape)), dtype=dtypes.int32)
    )
    w_affix = "" if w_arange is None else "_"
    y_fixed = y.symval if y_reshape is None else y_reshape
    y_affix = "" if y_reshape is None else "_"
    return f"""{f'''%{w.name}_i = "stablehlo.iota"() {{ iota_dimension = 0 : i64}} : {annotate_sig((), w_arange)}
%{w.name}_ = "stablehlo.concatenate"(%{w.name}_i, %{w.name}) {{ dimension = {b} : i64}} : {annotate_sig((w_arange, w.symval), w_fixed)}'''
    if w_arange is not None else ''
}%{y.name}{y_affix} = "stablehlo.gather"(%{x.name}, %{w.name}{w_affix}) {{
dimension_numbers = #stablehlo.gather<
offset_dims = {offset_dims},
collapsed_slice_dims = {collapsed_slice_dims},
start_index_map = {start_index_map},
index_vector_dim = {index_vector_dim}>,
slice_sizes = dense<{slice_sizes}> : tensor<{len(slice_sizes)}xi64>,
indices_are_sorted = false
}} : {annotate_sig((x.symval, w_fixed), y_fixed)}
{f'%{y.name} = "stablehlo.reshape"(%{y.name}_) : {annotate_sig((y_fixed,), y.symval)}' 
if y_reshape is not None else ''}"""


@backend.set_impl(backend.operator_set.scatter_nd)
def scatter_nd_impl(self, x, w, u, y):
    y_init_type = SymbolicTensor((), y.symval.dtype, y.symval.device)
    y_mlir_type = annotate_shape(y_init_type)

    r = x.symval.ndim
    q = w.symval.ndim
    s = u.symval.ndim
    # index_vector_dim = q - 1
    # update_window_dims = list(range(index_vector_dim, r))
    # inserted_window_dims = [0]
    # scatter_dims_to_operand_dims = [0]

    index_vector_dim = q - 1
    update_window_dims = list(range(index_vector_dim, r)) if q <= s else []
    # update_window_dims = []
    inserted_window_dims = list(range(r + 1 - s))
    scatter_dims_to_operand_dims = list(range(r + 1 - s))

    # if s <= r:
    #     index_vector_dim = 1
    #     update_window_dims = list(range(index_vector_dim, r))
    #     inserted_window_dims = list(range(r))
    #     scatter_dims_to_operand_dims = list(range(r))
    # else:
    #     index_vector_dim = q - 1
    #     update_window_dims = list(range(index_vector_dim, r))
    #     inserted_window_dims = [0]
    #     scatter_dims_to_operand_dims = [0]

    return f"""%{y.name} = "stablehlo.scatter"(%{x.name}, %{w.name}, %{u.name}) ({{
  ^bb0(%arg0: {y_mlir_type}, %arg1: {y_mlir_type}):
    %0 = "stablehlo.add"(%arg0, %arg1) : {annotate_sig((y_init_type, y_init_type), y_init_type)}
    "stablehlo.return"(%0) : ({y_mlir_type}) -> ()
}}) {{
  scatter_dimension_numbers = #stablehlo.scatter<
  update_window_dims = {update_window_dims},
  inserted_window_dims = {inserted_window_dims},
  scatter_dims_to_operand_dims = {scatter_dims_to_operand_dims},
  index_vector_dim = {index_vector_dim}>,
  indices_are_sorted = false,
  unique_indices = false
}} : {annotate_sig((x.symval, w.symval, u.symval), y.symval)}
"""
    #   update_window_dims = [1, 2],
    #   inserted_window_dims = [0],
    #   scatter_dims_to_operand_dims = [0],
    #   index_vector_dim = 1>,
    #   indices_are_sorted = false,
    #   unique_indices = false



import slope

y_matmat = slope.ones(4, 5) @ slope.ones(5, 6)
print(f"{y_matmat=}")

y_vecmat = slope.ones(5) @ slope.ones(5, 6)
print(f"{y_vecmat=}")

y_matvec = slope.ones(4, 5) @ slope.ones(5)
print(f"{y_matvec=}")

y_bmatvec = slope.ones(2, 4, 5) @ slope.ones(5)
print(f"{y_bmatvec=}")

y_bvecmat = slope.ones(5) @ slope.ones(2, 5, 6)
print(f"{y_bvecmat=}")



import slope
from slope import base_ops
from slope.tensor_shape import SymbolicTensor
import numpy as np

from typing import Union, List, Tuple, Sequence, Any, Callable, NamedTuple
from abc import ABC, abstractmethod

import numpy as np


class RandomBitGenerator(base_ops.Operator):
    @staticmethod
    def run(key, *, shape, dtype, algorithm):
        return []

    def typecheck(key, *, shape, dtype, algorithm):
        return [key.shape, key.shape]


def rng_bit_generator(key, shape, dtype=np.uint32, algorithm="default"):
    if np.dtype(dtype) not in {
        np.dtype("uint8"),
        np.dtype("uint16"),
        np.dtype("uint32"),
        np.dtype("uint64"),
    }:
        raise TypeError(f"rng_bit_generator: unsupported dtype {dtype}")
    return tuple(slope.RT.bind1(RandomBitGenerator, key, shape=shape, dtype=dtype, algorithm=algorithm))


class PRNGImpl(NamedTuple):
    key_shape: tuple
    seed: Callable
    split: Callable
    random_bits: Callable
    fold_in: Callable
    tag: str = "?"

    def __hash__(self) -> int:
        return hash(self.tag)

    def __str__(self) -> str:
        return self.tag


#   def pprint(self):
#     return (pp.text(f"{self.__class__.__name__} [{self.tag}]:") +
#             pp.nest(2, pp.group(pp.brk() + pp.join(pp.brk(), [
#               pp.text(f"{k} = {v}") for k, v in self._asdict().items()
#             ]))))


def _threefry_split_original(key, num):
    counts = base_ops.iota(np.uint32, num * 2)
    return base_ops.reshape(threefry_2x32(key, counts), (num, 2))


def threefry_seed(seed):
    """Create a single raw threefry PRNG key from an integer seed.

    Args:
      seed: a 64- or 32-bit integer used as the value of the key.

    Returns:
      The PRNG key contents, modeled as an tensor of shape (2,) and dtype
      uint32. The key is constructed from a 64-bit seed by effectively
      bit-casting to a pair of uint32 values (or from a 32-bit seed by
      first padding out with zeros).
    """
    if seed.shape:
        raise TypeError(f"PRNG key seed must be a scalar; got {seed!r}.")
    if not np.issubdtype(seed.dtype, np.integer):
        raise TypeError(f"PRNG key seed must be an integer; got {seed!r}")
    convert = lambda k: base_ops.reshape(convert(k, np.uint32), [1])
    k1 = convert(base_ops.shift_right_logical(seed, np.tensor(32, dtype=seed.dtype)))
    k2 = convert(np.bitwise_and(seed, np.uint32(0xFFFFFFFF)))
    return base_ops.cat([k1, k2], 0)


def random_seed(seeds, impl):
    seeds_arr = np.astensor(seeds)
    return random_seed(seeds_arr, impl=impl)


def PRNGKey(seed):
    if np.ndim(seed):
        raise TypeError(
            "PRNGKey accepts a scalar seed, but was given an tensor of"
            f"shape {np.shape(seed)} != (). Use jax.vmap for batching"
        )
    key = random_seed(seed, impl=rbg_prng_impl)
    if not isinstance(key, PRNGKeyTensorImpl):
        raise TypeError
    return random_unwrap(keys)


def _rbg_seed(seed: typing.Tensor) -> typing.Tensor:
    assert not seed.shape
    halfkey = threefry_seed(seed)
    return jnp.cat([halfkey, halfkey])


def _rbg_split(key, num: int):
    _threefry_split = _threefry_split_original
    return slope.ad.vmap(_threefry_split, (0, None), 1)(key.reshape(2, 2), num).reshape(num, 4)


def _rbg_fold_in(key, data):
    assert not data.shape
    return slope.ad.vmap(_threefry_fold_in, (0, None), 0)(key.reshape(2, 2), data).reshape(4)


def _rbg_random_bits(key, bit_width: int, shape: Sequence[int]):
    if not key.shape == (4,) and key.dtype == jnp.dtype("uint32"):
        raise TypeError("_rbg_random_bits got invalid prng key.")
    if bit_width not in (8, 16, 32, 64):
        raise TypeError("requires 8-, 16-, 32- or 64-bit field width.")
    _, bits = lax.rng_bit_generator(key, shape, dtype=UINT_DTYPES[bit_width])
    return bits


rbg_prng_impl = PRNGImpl(
    key_shape=(4,),
    seed=_rbg_seed,
    split=_rbg_split,
    random_bits=_rbg_random_bits,
    fold_in=_rbg_fold_in,
    tag="rbg",
)


def threefry_2x32(keypair, count):
    """Apply the Threefry 2x32 hash.

    Args:
      keypair: a pair of 32bit unsigned integers used for the key.
      count: an tensor of dtype uint32 used for the counts.

    Returns:
      An tensor of dtype uint32 with the same shape as `count`.
    """
    key1, key2 = keypair
    if not dtype(key1) == dtype(key2) == dtype(count) == np.uint32:
        msg = "threefry_2x32 requires uint32 arguments, got {}"
        raise TypeError(msg.format([dtype(x) for x in [key1, key2, count]]))

    assert count.size % 2

    if odd_size:
        x = list(jnp.split(jnp.cat([count.ravel(), np.uint32([0])]), 2))
    else:
        x = list(jnp.split(count.ravel(), 2))

    x = threefry2x32(key1, key2, x[0], x[1])
    out = cat(x)
    assert out.dtype == np.uint32
    return reshape(out[:-1] if odd_size else out, count.shape)


def threefry_split(key, num: int):
    counts = iota(np.uint32, num * 2)
    return reshape(threefry_2x32(key, counts), (num, 2))


def _threefry_split_foldlike(key, num):
    k1, k2 = key
    counts1, counts2 = iota_2x32_shape((num,))
    bits1, bits2 = threefry2x32(k1, k2, counts1, counts2)
    return stack([bits1, bits2], dim=1)


def threefry_fold_in(key, data):
    assert not data.shape
    return _threefry_fold_in(key, np.uint32(data))


def _threefry_fold_in(key, data):
    return threefry_2x32(key, threefry_seed(data))


#


class PRNGKeyTensorImpl(PRNGKeyTensor):
    """An tensor of PRNG keys backed by an RNG implementation.

    This class lifts the definition of a PRNG, provided in the form of a
    ``PRNGImpl``, into an tensor-like tree class. Instances of this
    class behave like an tensor whose base elements are keys, hiding the
    fact that keys are typically tensors (of ``uint32`` dtype) themselves.

    PRNGKeyTensors are also restricted relative to JAX tensors in that
    they do not expose arithmetic operations. They instead expose
    wrapper methods around the PRNG implementation functions (``split``,
    ``random_bits``, ``fold_in``).
    """

    impl: PRNGImpl
    _base_tensor: typing.Tensor

    def __init__(self, impl, key_data: Any):
        assert not isinstance(key_data, core.TraceTensor)
        _check_prng_key_data(impl, key_data)
        self.impl = impl
        self._base_tensor = key_data

    # TODO(frostig): rename to unsafe_base_tensor, or just offer base_tensor attr?
    def unsafe_raw_tensor(self):
        """Access the raw numerical tensor that carries underlying key data.

        Returns:
          A uint32 JAX tensor whose leading dimensions are ``self.shape``.
        """
        return self._base_tensor

    def block_until_ready(self):
        _ = self._base_tensor.block_until_ready()
        return self

    @property
    def shape(self):
        return base_arr_shape_to_keys_shape(self.impl, self._base_tensor.shape)

    @property
    def ndim(self):
        return len(self.shape)

    @property
    def dtype(self):
        return KeyTy(self.impl)

    _device = property(op.attrgetter("_base_tensor._device"))
    _committed = property(op.attrgetter("_base_tensor._committed"))

    @property
    def sharding(self):
        symtensor = keys_shaped_tensor(self.impl, self.shape)
        phys_sharding = self._base_tensor.sharding
        return KeyTyRules.logical_op_sharding(symtensor, phys_sharding)

    def _is_scalar(self):
        base_ndim = len(self.impl.key_shape)
        return self._base_tensor.ndim == base_ndim

    def __len__(self):
        if self._is_scalar():
            raise TypeError("len() of unsized object")
        return len(self._base_tensor)

    def __iter__(self) -> Iterator[PRNGKeyTensorImpl]:
        if self._is_scalar():
            raise TypeError("iteration over a 0-d key tensor")
        # TODO(frostig): we may want to avoid iteration by slicing because
        # a very common use of iteration is `k1, k2 = split(key)`, and
        # slicing/indexing may be trickier to track for linearity checking
        # purposes. Maybe we can:
        # * introduce an unpack primitive+traceable (also allow direct use)
        # * unpack upfront into shape[0] many keytensor slices
        # * return iter over these unpacked slices
        # Whatever we do, we'll want to do it by overriding
        # ShapedTensor._iter when the element type is KeyTy...
        return (PRNGKeyTensorImpl(self.impl, k) for k in iter(self._base_tensor))

    # TODO(frostig): are all of the stackable methods below (reshape,
    # concat, expand, expand_dims), and the stackable registration,
    # still needed? If, with some work, none are needed, then do we want
    # to remove stackables altogether? This may be the only application.

    # TODO(frostig): Remove? Overwritten below in particular
    def reshape(self, newshape, order=None) -> PRNGKeyTensorImpl:
        reshaped_base = jnp.reshape(self._base_tensor, (*newshape, -1), order=order)
        return PRNGKeyTensorImpl(self.impl, reshaped_base)

    def cat(self, key_arrs, dim, dtype=None) -> PRNGKeyTensorImpl:
        if dtype is not None:
            raise ValueError("dtype argument not supported for concatenating PRNGKeyTensor")
        dim = canonicalize_dim(dim, self.ndim)
        arrs = [self._base_tensor, *[k._base_tensor for k in key_arrs]]
        return PRNGKeyTensorImpl(self.impl, jnp.cat(arrs, dim))

    def expand(self, shape) -> PRNGKeyTensorImpl:
        if jnp.ndim(shape) == 0:
            shape = (shape,)
        new_shape = (*shape, *self.impl.key_shape)
        return PRNGKeyTensorImpl(self.impl, jnp.expand(self._base_tensor, new_shape))

    def expand_dims(self, dimensions: Sequence[int]) -> PRNGKeyTensorImpl:
        # follows lax.expand_dims, not jnp.expand_dims, so dimensions is a sequence
        ndim_out = self.ndim + len(set(dimensions))
        dimensions = [canonicalize_dim(d, ndim_out) for d in dimensions]
        return PRNGKeyTensorImpl(self.impl, lax.expand_dims(self._base_tensor, dimensions))

    def __repr__(self):
        return f"{self.__class__.__name__}[{self.impl.tag}]" f" {{ {self._base_tensor} }}"

    def pprint(self):
        pp_keys = pp.text("shape = ") + pp.text(str(self.shape))
        pp_impl = pp.text("impl = ") + self.impl.pprint()
        return str(pp.group(pp.text("PRNGKeyTensor:") + pp.nest(2, pp.brk() + pp_keys + pp.brk() + pp_impl)))

    # Overwritten immediately below
    @property
    def T(self) -> PRNGKeyTensor:
        assert False

    def __getitem__(self, _) -> PRNGKeyTensor:
        assert False

    def ravel(self, *_, **__) -> PRNGKeyTensor:
        assert False

    def squeeze(self, *_, **__) -> PRNGKeyTensor:
        assert False

    def transpose(self, *_, **__) -> PRNGKeyTensor:
        assert False

    def take(self, *_, **__) -> PRNGKeyTensor:
        assert False

    def permute(self, *_, **__) -> PRNGKeyTensor:
        assert False

    def flatten(self, *_, **__) -> PRNGKeyTensor:
        assert False


_set_device_tensor_base_attributes(
    PRNGKeyTensorImpl,
    include=[
        "__getitem__",
        "ravel",
        "squeeze",
        "transpose",
        "take",
        "reshape",
        "permute",
        "flatten",
        "T",
    ],
)
_register_stackable(PRNGKeyTensorImpl)
basetensor.Tensor.register(PRNGKeyTensorImpl)


# TODO(frostig): remove, rerouting callers directly to random_seed
def seed_with_impl(impl: PRNGImpl, seed: Union[int, Tensor]) -> PRNGKeyTensorImpl:
    return random_seed(seed, impl=impl)


def keys_shaped_tensor(impl, shape):
    return core.ShapedTensor(shape, KeyTy(impl))


def keys_symtensor_to_base_arr_symtensor(keys_symtensor):
    shape = (*keys_symtensor.shape, *keys_symtensor.dtype.impl.key_shape)
    return core.ShapedTensor(shape, np.dtype("uint32"))


def base_arr_shape_to_keys_shape(impl, base_arr_shape):
    base_ndim = len(impl.key_shape)
    return base_arr_shape[:-base_ndim]


def make_key_tensor_phys_sharding(symtensor, sharding, is_sharding_from_xla):
    if dispatch.is_single_device_sharding(sharding):
        return sharding
    elif isinstance(sharding, PmapSharding):
        key_shape = symtensor.dtype.impl.key_shape
        trailing_sharding = [sharding_specs.NoSharding()] * len(key_shape)
        phys_sharding_spec = sharding_specs.ShardingSpec(
            sharding=(*sharding.sharding_spec.sharding, *trailing_sharding),
            mesh_mapping=sharding.sharding_spec.mesh_mapping,
        )
        return PmapSharding(devices=sharding.devices, sharding_spec=phys_sharding_spec)
    elif isinstance(sharding, NamedSharding):
        key_shape = symtensor.dtype.impl.key_shape
        trailing_spec = [None] * len(key_shape)
        return NamedSharding(sharding.mesh, PartitionSpec(*sharding.spec, *trailing_spec))
    elif is_sharding_from_xla:
        return sharding
    else:
        sharding_proto = sharding._to_xla_op_sharding(symtensor.ndim)
        return GSPMDSharding(
            sharding._device_assignment,
            KeyTyRules.physical_op_sharding(symtensor, sharding_proto),
        )


##############


def philox(cls, key, rounds=10):
    def rotl(x, r):
        return (x << r) | (x >> (64 - r))

    def philox_round(a, b, key):
        a += b
        b = rotl(b, 32)
        a ^= key
        return a, b

    assert len(key) == 2, "Key must be a tuple of two 64-bit integers"
    assert isinstance(rounds, int) and rounds > 0, "Number of rounds must be a positive integer"

    key0, key1 = key
    state = np.tensor([key0, key1], dtype=np.uint64)

    for _ in range(rounds):
        state[0], state[1] = philox_round(state[0], state[1], _)
        state[0] += 1
    return state[:2], state  #  [0, 1] are new keys, [:] are random numbers,


def threefry(cls, key, rounds=20):
    def rotl(x, r):
        return (x << r) | (x >> (64 - r))

    def threefry_round(a, b, c, d, r):
        a += b
        d ^= a
        d = rotl(d, r)
        c += d
        b ^= c
        b = rotl(b, r)
        a += b
        d ^= a
        d = rotl(d, r)
        c += d
        b ^= c
        b = rotl(b, r)
        return a, b, c, d

    assert len(key) == 2, "Key must be a tuple of two 64-bit integers"
    assert isinstance(rounds, int) and rounds > 0, "Number of rounds must be a positive integer"

    key0, key1 = key
    state = np.tensor([0, 0, key0, key1], dtype=np.uint64)

    for _ in range(rounds):
        state[0], state[1], state[2], state[3] = threefry_round(state[0], state[1], state[2], state[3], 14)
        state[0], state[1], state[2], state[3] = threefry_round(state[0], state[1], state[2], state[3], 16)
        state[0], state[1], state[2], state[3] = threefry_round(state[0], state[1], state[2], state[3], 52)
        state[0], state[1], state[2], state[3] = threefry_round(state[0], state[1], state[2], state[3], 57)

    return state[:2], state[2:]  # [0, 1] are new keys, [2:] are random numbers


def rng_bit(cls, key, algorithm="philox"):
    return dict(philox=cls.philox, threefry=cls.threefry)[algorithm](key)


def random_normal(cls, x, key, dtype):
    # Box-Muller transform
    nbits = dtype.itemsize * 8
    u1 = 0
    while u1 == 0:
        u1, key = cls.rng_bit(x, key)
        u1 = u1 / (2**nbits)  # normalize to [0, 1]
    u2, key = cls.rng_bit(x, key)
    u2 = u2 / (2**nbits)
    z0 = (-2.0 * u1.log()).sqrt() * (2 * math.pi * u2).cos()
    return z0, key



import slope
from slope import nn


class Linear(nn.Module):
    def __init__(self, in_dim, out_dim, bias=False):
        self.weight = slope.randn((out_dim, in_dim))
        self.bias = slope.zeros(out_dim) if bias else None

    def __call__(self, x):
        x = x @ self.weight.transpose(-1, -2)
        return x + self.bias if self.bias is not None else x


def loss_fn(model, batch):
    x, y = batch
    y_hat = model(x)
    loss = (y - y_hat).sum()
    return loss


g_loss_fn = slope.grad(loss_fn)


@slope.jit
def train_step(model, batch, optim):
    g_model = g_loss_fn(model, batch)
    model, optim = optim(model, g_model)
    return model, optim


x = slope.ones((1, 2))
y = slope.ones((1, 1))
model = Linear(2, 3, 1)
optim = nn.SGD(model)
model, optim = train_step(model, (x, y), optim)
print(model.flatten())
print(optim.flatten())



import slope

x1 = slope.tensor([1.0])
x2 = slope.tensor([2.0])
# x1 = x2 = slope.full((), 1.)

y1 = x1 + x2
print(f"{y1=}")
y2 = x1 + x2
print(f"{y2=}")


# x3 = slope.ones(2,3)
# y3 = x3.sum(0, keepdim=True)
# y3 = x3.max(0, keepdim=True)
# print(f"{y3=}")

# x4 = slope.tensor([1.,2.,3.])
# y4 = x4.max(0, keepdim=True)
# print(f"{y4=}")



import slope
import slope.nn as nn
import math

x = slope.ones((1, 2))
y = slope.full((1,), 1)
model = nn.Linear(2, 1)


@slope.jit
def loss_fn(model, batch):
    x, y = batch
    y_hat = model(x)
    loss = (y - y_hat).pow(2).sum()
    return loss


L = loss_fn(model, (x, y))
gloss_fn = slope.grad(loss_fn)
gL = gloss_fn(model, (x, y)).flatten()

print(f"{L=}, {gL=}")
